from .rosenbrock import *
from .fmin import *
from .fmax import *
from .is_symbolic import *
from .is_numeric import *
from .sign import *
from .norm2 import *
from .fabs import *
from .sq_dist_to_soc import *
from .norm2_squared import *



from .constraint import Constraint
import numpy as np


class Simplex(Constraint):
    """Simplex constraint

    A set of the form {x: sum(x) = a, x>=0} where alpha is a positive constant.

    """

    def __init__(self, alpha: float = 1.0):  # unless specified, alpha=1
        """Constructor for a Simplex constraint

        Args:
            :param alpha: size parameter of simplex (default: 1)

        Returns:
            :return: new instance of Simplex with given alpha
        """
        if alpha <= 0:
            raise Exception("Alpha must be a positive number")

        self.__alpha = float(alpha)

    @property
    def alpha(self):
        """Returns the simplex value alpha"""
        return self.__alpha

    def distance_squared(self, u):
        raise NotImplementedError()

    def project(self, y):
        """Computes the projection of a given point `y` (in R^n) on the current simplex.

        Args:
            :param y: given point; must be a list of numbers (float, int) or
            a numpy n-dim array (`ndarray`)

        Returns:
            :return: the projection point in R^n as a numpy array of float64s
        """
        def __pop_all(z, indices):
            for index in sorted(indices, reverse=True):
                del z[index]
            return z

        a = self.__alpha

        # 1 ----
        v = [y[0]]
        v_size_old = -1
        v_tilde = []
        rho = y[0] - a

        # 2 ----
        for yn in y[1:]:
            if yn > rho:
                rho += (yn - rho) / (len(v) + 1)
                if rho > yn - a:
                    v.append(yn)
                else:
                    v_tilde.extend(v)
                    v = [yn]
                    rho = yn - a

        # 3 ----
        if len(v_tilde) > 0:
            for v_tilde_i in v_tilde:
                if v_tilde_i > rho:
                    v.append(v_tilde_i)
                    rho += (v_tilde_i - rho) / len(v)

        # 4 ----
        keep_running = True
        while keep_running:
            hit_list = []
            current_len_v = len(v)
            for j, vj in enumerate(v):
                if vj <= rho:
                    hit_list += [j]
                    current_len_v -= 1
                    rho += (rho - vj) / current_len_v
            v = __pop_all(v, hit_list)
            keep_running = current_len_v != v_size_old
            v_size_old = current_len_v

        # 6 ----
        ufunc = np.vectorize(lambda s: max(s - rho, 0), otypes=[np.float64])
        x = ufunc(np.array(y, dtype=np.float64))

        # result ----
        return x

    def is_convex(self):
        return True

    def is_compact(self):
        return True



import os
import unittest
import casadi.casadi as cs
import opengen as og
import subprocess
import logging


class RustBuildTestCase(unittest.TestCase):

    TEST_DIR = ".python_test_build"

    @staticmethod
    def get_open_local_absolute_path():
        cwd = os.getcwd()
        return cwd.split('open-codegen')[0]

    # Which version of OpEn Rust library to test against
    OPEN_RUSTLIB_VERSION = "*"

    @classmethod
    def solverConfig(cls):
        solver_config = og.config.SolverConfiguration() \
            .with_lbfgs_memory(15) \
            .with_tolerance(1e-4) \
            .with_initial_tolerance(1e-4) \
            .with_delta_tolerance(1e-4) \
            .with_initial_penalty(15.0) \
            .with_penalty_weight_update_factor(10.0) \
            .with_max_inner_iterations(155) \
            .with_max_duration_micros(1e8) \
            .with_max_outer_iterations(50) \
            .with_sufficient_decrease_coefficient(0.05) \
            .with_cbfgs_parameters(1.5, 1e-10, 1e-12)
        return solver_config

    @classmethod
    def setUpPythonBindings(cls):
        u = cs.MX.sym("u", 5)  # decision variable (nu = 5)
        p = cs.MX.sym("p", 2)  # parameter (np = 2)
        phi = og.functions.rosenbrock(u, p)  # cost function
        bounds = og.constraints.Ball2(None, 1.5)  # ball centered at origin
        meta = og.config.OptimizerMeta() \
            .with_optimizer_name("python_bindings")
        problem = og.builder.Problem(u, p, phi) \
            .with_constraints(bounds)
        build_config = og.config.BuildConfiguration() \
            .with_open_version(RustBuildTestCase.OPEN_RUSTLIB_VERSION) \
            .with_build_directory(RustBuildTestCase.TEST_DIR) \
            .with_build_mode(og.config.BuildConfiguration.DEBUG_MODE)\
            .with_build_python_bindings()
        og.builder.OpEnOptimizerBuilder(problem,
                                        metadata=meta,
                                        build_configuration=build_config,
                                        solver_configuration=cls.solverConfig()) \
            .build()

    @classmethod
    def setUpOnlyF1(cls):
        u = cs.MX.sym("u", 5)  # decision variable (nu = 5)
        p = cs.MX.sym("p", 2)  # parameter (np = 2)
        f1 = cs.vertcat(1.5 * u[0] - u[1], u[2] - u[3])
        set_c = og.constraints.Rectangle(
            xmin=[-0.01, -0.01], xmax=[0.02, 0.03])
        phi = og.functions.rosenbrock(u, p)  # cost function
        bounds = og.constraints.Ball2(None, 1.5)  # ball centered at origin
        tcp_config = og.config.TcpServerConfiguration(bind_port=3301)
        meta = og.config.OptimizerMeta() \
            .with_optimizer_name("only_f1")
        problem = og.builder.Problem(u, p, phi) \
            .with_aug_lagrangian_constraints(f1, set_c) \
            .with_constraints(bounds)
        build_config = og.config.BuildConfiguration() \
            .with_open_version(local_path=RustBuildTestCase.get_open_local_absolute_path())  \
            .with_build_directory(RustBuildTestCase.TEST_DIR) \
            .with_build_mode(og.config.BuildConfiguration.DEBUG_MODE) \
            .with_tcp_interface_config(tcp_interface_config=tcp_config) \
            .with_build_c_bindings()  \
            .with_allocator(og.config.RustAllocator.JemAlloc)
        og.builder.OpEnOptimizerBuilder(problem,
                                        metadata=meta,
                                        build_configuration=build_config,
                                        solver_configuration=cls.solverConfig()) \
            .build()

    @classmethod
    def setUpOnlyF2(cls):
        u = cs.MX.sym("u", 5)  # decision variable (nu = 5)
        p = cs.MX.sym("p", 2)  # parameter (np = 2)
        f2 = cs.vertcat(0.2 + 1.5 * u[0] - u[1], u[2] - u[3] - 0.1)
        phi = og.functions.rosenbrock(u, p)
        bounds = og.constraints.Ball2(None, 1.5)
        tcp_config = og.config.TcpServerConfiguration(bind_port=3302)
        meta = og.config.OptimizerMeta() \
            .with_optimizer_name("only_f2")
        problem = og.builder.Problem(u, p, phi) \
            .with_penalty_constraints(f2) \
            .with_constraints(bounds)
        build_config = og.config.BuildConfiguration() \
            .with_open_version(RustBuildTestCase.OPEN_RUSTLIB_VERSION)  \
            .with_build_directory(RustBuildTestCase.TEST_DIR) \
            .with_build_mode(og.config.BuildConfiguration.DEBUG_MODE) \
            .with_tcp_interface_config(tcp_interface_config=tcp_config) \
            .with_build_c_bindings()
        og.builder.OpEnOptimizerBuilder(problem,
                                        metadata=meta,
                                        build_configuration=build_config,
                                        solver_configuration=cls.solverConfig()) \
            .build()

    @classmethod
    def setUpPlain(cls):
        u = cs.MX.sym("u", 5)  # decision variable (nu = 5)
        p = cs.MX.sym("p", 2)  # parameter (np = 2)
        phi = og.functions.rosenbrock(u, p)
        bounds = og.constraints.Ball2(None, 1.5)
        tcp_config = og.config.TcpServerConfiguration(bind_port=4598)
        meta = og.config.OptimizerMeta() \
            .with_optimizer_name("plain")
        problem = og.builder.Problem(u, p, phi) \
            .with_constraints(bounds)
        build_config = og.config.BuildConfiguration() \
            .with_open_version(RustBuildTestCase.OPEN_RUSTLIB_VERSION)  \
            .with_build_directory(RustBuildTestCase.TEST_DIR) \
            .with_build_mode(og.config.BuildConfiguration.DEBUG_MODE) \
            .with_tcp_interface_config(tcp_interface_config=tcp_config) \
            .with_build_c_bindings()
        og.builder.OpEnOptimizerBuilder(problem,
                                        metadata=meta,
                                        build_configuration=build_config,
                                        solver_configuration=cls.solverConfig()) \
            .build()

    @classmethod
    def setUpRosPackageGeneration(cls):
        u = cs.MX.sym("u", 5)  # decision variable (nu = 5)
        p = cs.MX.sym("p", 2)  # parameter (np = 2)
        phi = og.functions.rosenbrock(u, p)
        c = cs.vertcat(1.5 * u[0] - u[1],
                       cs.fmax(0.0, u[2] - u[3] + 0.1))
        bounds = og.constraints.Ball2(None, 1.5)
        meta = og.config.OptimizerMeta() \
            .with_optimizer_name("rosenbrock_ros")
        problem = og.builder.Problem(u, p, phi) \
            .with_constraints(bounds) \
            .with_penalty_constraints(c)
        ros_config = og.config.RosConfiguration() \
            .with_package_name("parametric_optimizer") \
            .with_node_name("open_node") \
            .with_rate(35) \
            .with_description("really cool ROS node")
        build_config = og.config.BuildConfiguration() \
            .with_open_version(RustBuildTestCase.OPEN_RUSTLIB_VERSION)  \
            .with_build_directory(RustBuildTestCase.TEST_DIR) \
            .with_build_mode(og.config.BuildConfiguration.DEBUG_MODE) \
            .with_build_c_bindings()  \
            .with_ros(ros_config)
        og.builder.OpEnOptimizerBuilder(problem,
                                        metadata=meta,
                                        build_configuration=build_config,
                                        solver_configuration=cls.solverConfig()) \
            .build()

    @classmethod
    def setUpOnlyParametricF2(cls):
        u = cs.MX.sym("u", 5)  # decision variable (nu = 5)
        p = cs.MX.sym("p", 3)  # parameter (np = 3)
        f2 = u[0] - p[2]
        phi = og.functions.rosenbrock(
            u, cs.vertcat(p[0], p[1]))  # cost function
        bounds = og.constraints.Ball2(None, 1.5)  # ball centered at origin
        tcp_config = og.config.TcpServerConfiguration(bind_port=4599)
        meta = og.config.OptimizerMeta() \
            .with_optimizer_name("parametric_f2")
        problem = og.builder.Problem(u, p, phi) \
            .with_penalty_constraints(f2) \
            .with_constraints(bounds)
        build_config = og.config.BuildConfiguration() \
            .with_open_version(RustBuildTestCase.OPEN_RUSTLIB_VERSION) \
            .with_build_mode(og.config.BuildConfiguration.DEBUG_MODE) \
            .with_build_directory(RustBuildTestCase.TEST_DIR) \
            .with_tcp_interface_config(tcp_interface_config=tcp_config) \
            .with_build_c_bindings()
        solver_config = og.config.SolverConfiguration() \
            .with_tolerance(1e-6) \
            .with_initial_tolerance(1e-4) \
            .with_delta_tolerance(1e-5)
        og.builder.OpEnOptimizerBuilder(
            problem, meta, build_config, solver_config).build()

    @classmethod
    def setUpHalfspace(cls):
        u = cs.SX.sym("u", 5)  # decision variable (nu = 5)
        p = cs.SX.sym("p", 2)  # parameter (np = 2)
        phi = cs.dot(u, u)  # cost function

        bounds = og.constraints.Halfspace([1., 2., 1., 5., 2.], -10.39)

        problem = og.builder.Problem(u, p, phi) \
            .with_constraints(bounds)

        meta = og.config.OptimizerMeta() \
            .with_optimizer_name("halfspace_optimizer")

        tcp_config = og.config.TcpServerConfiguration(bind_port=3305)
        build_config = og.config.BuildConfiguration() \
            .with_build_directory(RustBuildTestCase.TEST_DIR) \
            .with_build_mode(og.config.BuildConfiguration.DEBUG_MODE) \
            .with_open_version(RustBuildTestCase.OPEN_RUSTLIB_VERSION) \
            .with_tcp_interface_config(tcp_interface_config=tcp_config)

        builder = og.builder.OpEnOptimizerBuilder(problem,
                                                  meta,
                                                  build_config,
                                                  cls.solverConfig())
        builder.build()

    @classmethod
    def setUpClass(cls):
        cls.setUpPythonBindings()
        cls.setUpRosPackageGeneration()
        cls.setUpOnlyF1()
        cls.setUpOnlyF2()
        cls.setUpPlain()
        cls.setUpOnlyParametricF2()
        cls.setUpHalfspace()

    def test_python_bindings(self):
        import sys
        import os

        # include the target directory into the path...
        sys.path.insert(1, os.path.join(
            RustBuildTestCase.TEST_DIR, "python_bindings"))
        import python_bindings  # import python_bindings.so

        solver = python_bindings.solver()
        # returns object of type OptimizerSolution
        result = solver.run([1., 2.])
        self.assertIsNotNone(result.solution)

    def test_rectangle_empty(self):
        xmin = [-1, 2]
        xmax = [-2, 4]
        with self.assertRaises(Exception) as __context:
            og.constraints.Rectangle(xmin, xmax)

    def test_rectangle_incompatible_dimensions(self):
        xmin = [-1, -1, 1]
        xmax = [1, 1]
        with self.assertRaises(Exception) as __context:
            og.constraints.Rectangle(xmin, xmax)

    def test_rectangle_both_none(self):
        with self.assertRaises(Exception) as __context:
            og.constraints.Rectangle(None, None)

    def test_ball_negative_radius(self):
        with self.assertRaises(Exception) as __context:
            og.constraints.Ball2(None, -1)

    def test_solver_config_wrong_max_duration(self):
        with self.assertRaises(Exception) as __context:
            og.config.SolverConfiguration().with_max_duration_micros(0)

    def test_solver_config_wrong_update_factor(self):
        with self.assertRaises(Exception) as __context:
            og.config.SolverConfiguration().with_penalty_weight_update_factor(0.5)

    def test_solver_config_wrong_outer_iterations(self):
        with self.assertRaises(Exception) as __context:
            og.config.SolverConfiguration().with_max_outer_iterations(0)

    def test_solver_config_wrong_inner_iterations(self):
        with self.assertRaises(Exception) as __context:
            og.config.SolverConfiguration().with_max_inner_iterations(0)

    def test_solver_config_wrong_constraints_tolerance(self):
        with self.assertRaises(Exception) as __context:
            og.config.SolverConfiguration().with_delta_tolerance(0)

    def test_solver_config_wrong_inner_tolerance(self):
        with self.assertRaises(Exception) as __context:
            og.config.SolverConfiguration().with_tolerance(0)

    def test_solver_config_wrong_lbfgs_memory(self):
        with self.assertRaises(Exception) as __context:
            og.config.SolverConfiguration().with_lbfgs_memory(1)

    def test_solver_config_wrong_max_inner_iterations(self):
        with self.assertRaises(Exception) as __context:
            og.config.SolverConfiguration().with_max_inner_iterations()

    def test_start_multiple_servers(self):
        all_managers = []
        for i in range(50):
            all_managers += [og.tcp.OptimizerTcpManager(
                optimizer_path=RustBuildTestCase.TEST_DIR + '/only_f1',
                ip='0.0.0.0',
                port=15311+i)]

        # Start all servers
        for m in all_managers:
            m.start()

        # Ping all
        for m in all_managers:
            m.ping()

        # Kill all
        for m in all_managers:
            m.kill()

    def test_rust_build_only_f1(self):
        # Start the server using a custom bind IP and port
        mng = og.tcp.OptimizerTcpManager(RustBuildTestCase.TEST_DIR + '/only_f1',
                                         ip='0.0.0.0',
                                         port=13757)
        mng.start()
        pong = mng.ping()  # check if the server is alive
        self.assertEqual(1, pong["Pong"])

        # Regular call
        response = mng.call(p=[2.0, 10.0])
        self.assertEqual("Converged", response["exit_status"])

        # Call with initial params, initial y and initial penalty param
        response = mng.call(p=[2.0, 10.0],
                            initial_guess=response["solution"],
                            initial_y=response["lagrange_multipliers"],
                            initial_penalty=response["penalty"])
        self.assertTrue(response.is_ok())
        status = response.get()
        self.assertEqual(2, status.num_outer_iterations)

        response = mng.call(p=[2.0, 10.0, 50.0])
        status = response.get()
        self.assertFalse(response.is_ok())
        self.assertEqual(True, isinstance(status, og.tcp.SolverError))
        self.assertEqual(3003, status.code)

        response = mng.call(p=[2.0, 10.0], initial_guess=[0.1, 0.2])
        self.assertFalse(response.is_ok())
        status = response.get()
        self.assertEqual(True, isinstance(status, og.tcp.SolverError))
        self.assertEqual(1600, status.code)

        response = mng.call(p=[2.0, 10.0], initial_y=[0.1])
        status = response.get()
        self.assertFalse(response.is_ok())
        self.assertEqual(True, isinstance(status, og.tcp.SolverError))
        self.assertEqual(1700, status.code)

        mng.kill()

    def test_rust_build_only_f2(self):
        mng = og.tcp.OptimizerTcpManager(
            RustBuildTestCase.TEST_DIR + '/only_f2')
        mng.start()
        pong = mng.ping()  # check if the server is alive
        self.assertEqual(1, pong["Pong"])

        # Regular call
        response = mng.call(p=[2.0, 10.0])
        status = response.get()
        self.assertEqual("Converged", status.exit_status)

        # Call with initial params, initial y and initial penalty param
        response = mng.call(p=[2.0, 10.0],
                            initial_guess=response["solution"],
                            initial_penalty=response["penalty"])
        self.assertTrue(response.is_ok())
        status = response.get()
        self.assertEqual(1, status.num_outer_iterations)

        response = mng.call(p=[2.0, 10.0, 50.0])
        self.assertFalse(response.is_ok())
        status = response.get()
        self.assertEqual(True, isinstance(status, og.tcp.SolverError))
        self.assertEqual(3003, status.code)

        response = mng.call(p=[2.0, 10.0], initial_guess=[0.1, 0.2])
        self.assertFalse(response.is_ok())
        status = response.get()
        self.assertEqual(True, isinstance(status, og.tcp.SolverError))
        self.assertEqual(1600, status.code)

        response = mng.call(p=[2.0, 10.0], initial_y=[0.1])
        self.assertFalse(response.is_ok())
        status = response.get()
        self.assertEqual(True, isinstance(status, og.tcp.SolverError))
        self.assertEqual(1700, status.code)

        mng.kill()

    def test_rust_build_plain(self):
        mng = og.tcp.OptimizerTcpManager(RustBuildTestCase.TEST_DIR + '/plain')
        mng.start()
        pong = mng.ping()  # check if the server is alive
        self.assertEqual(1, pong["Pong"])

        # Regular call
        response = mng.call(p=[2.0, 10.0])
        self.assertTrue(response.is_ok())
        status = response.get()
        self.assertEqual("Converged", status.exit_status)

        mng.kill()

    def test_rust_build_parametric_f2(self):
        # introduced to tackle issue #123
        mng = og.tcp.OptimizerTcpManager(
            RustBuildTestCase.TEST_DIR + '/parametric_f2')
        mng.start()
        pong = mng.ping()  # check if the server is alive
        self.assertEqual(1, pong["Pong"])

        # Regular call
        response = mng.call(p=[1.0, 1.0, 0.5])
        self.assertTrue(response.is_ok())
        status = response.get()
        self.assertEqual("Converged", status.exit_status)
        self.assertTrue(status.f2_norm < 1e-4)
        mng.kill()

    def test_rust_build_parametric_halfspace(self):
        mng = og.tcp.OptimizerTcpManager(
            RustBuildTestCase.TEST_DIR + '/halfspace_optimizer')
        mng.start()
        pong = mng.ping()  # check if the server is alive
        self.assertEqual(1, pong["Pong"])

        # Regular call
        response = mng.call(p=[1.0, 1.0])
        self.assertTrue(response.is_ok())
        status = response.get()
        self.assertEqual("Converged", status.exit_status)
        u = status.solution
        c = [1., 2., 1., 5., 2.]
        b = -10.39
        eps = 1e-14
        self.assertTrue(sum([u[i] * c[i] for i in range(5)]) - b <= eps)
        self.assertTrue(-sum([u[i] * c[i] for i in range(5)]) + b <= eps)

        mng.kill()

    @staticmethod
    def c_bindings_helper(optimizer_name):
        p = subprocess.Popen(["/usr/bin/gcc",
                              RustBuildTestCase.TEST_DIR + "/" + optimizer_name + "/example_optimizer.c",
                              "-I" + RustBuildTestCase.TEST_DIR + "/" + optimizer_name,
                              "-pthread",
                              RustBuildTestCase.TEST_DIR + "/" + optimizer_name +
                              "/target/debug/lib" + optimizer_name + ".a",
                              "-lm",
                              "-ldl",
                              "-std=c99",
                              "-o",
                              RustBuildTestCase.TEST_DIR + "/" + optimizer_name + "/optimizer"],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)

        # Make sure it compiles
        p.communicate()
        rc1 = p.returncode

        # Run the optimizer
        p = subprocess.Popen([RustBuildTestCase.TEST_DIR + "/" + optimizer_name + "/optimizer"],
                             stdout=subprocess.DEVNULL)
        p.communicate()
        rc2 = p.returncode

        return rc1, rc2

    def test_c_bindings(self):
        rc1, rc2 = RustBuildTestCase.c_bindings_helper(
            optimizer_name="only_f1")
        self.assertEqual(0, rc1)
        self.assertEqual(0, rc2)

        rc1, rc2 = RustBuildTestCase.c_bindings_helper(
            optimizer_name="only_f2")
        self.assertEqual(0, rc1)
        self.assertEqual(0, rc2)

        rc1, rc2 = RustBuildTestCase.c_bindings_helper(optimizer_name="plain")
        self.assertEqual(0, rc1)
        self.assertEqual(0, rc2)

    def test_tcp_manager_remote_cannot_start(self):
        remote_tcp_manager = og.tcp.OptimizerTcpManager(
            ip='10.8.0.1', port=3345)
        with self.assertRaises(Exception) as __context:
            remote_tcp_manager.start()

    def test_tcp_manager_remote_ip_no_port(self):
        with self.assertRaises(Exception) as __context:
            _remote_tcp_manager = og.tcp.OptimizerTcpManager(ip='10.8.0.1')

    def test_tcp_manager_remote_port_no_ip(self):
        with self.assertRaises(Exception) as __context:
            _remote_tcp_manager = og.tcp.OptimizerTcpManager(port=8888)


if __name__ == '__main__':
    logging.getLogger('retry').setLevel(logging.ERROR)
    unittest.main()



from opengen.config.tcp_server_config import TcpServerConfiguration
from opengen.config.ros_config import RosConfiguration
import random
import string
from enum import Enum


class RustAllocator(Enum):
    """Memory Allocator for the auto-generated project"""

    DefaultAllocator = 0
    """Default allocator"""

    JemAlloc = 1
    """Memory allocator: jemalloc
    
    Jemalloc is a generic implementation of malloc that emphasises fragmentation
    avoidance
    """

    RpAlloc = 2
    """Memory allocator: rpmalloc
    
    Rpmalloc is a very efficient lock-free thread caching 16-byte aligned 
    memory allocator implemented in C.
    """


class BuildConfiguration:
    """Build configuration

    Configuration for the code generator

    """

    DEBUG_MODE = "debug"
    RELEASE_MODE = "release"

    def __init__(self, build_dir="."):
        """
        Construct an instance of BuildConfiguration

        :param build_dir: Target directory, defaults to the current directory

        :return: A new instance of BuildConfiguration

        """
        random_string = ''.join(random.choice(
            string.ascii_letters) for _i in range(20))

        self.__target_system = None
        self.__build_mode = BuildConfiguration.RELEASE_MODE
        self.__id = random_string
        self.__cost_function_name = 'phi_' + random_string
        self.__grad_cost_function_name = 'grad_phi_' + random_string
        self.__constraint_penalty_function = 'mapping_f2_' + random_string
        self.__alm_constraints_mapping_f1 = 'mapping_f1_' + random_string
        self.__rebuild = False
        self.__build_dir = build_dir
        self.__open_version = None
        self.__build_c_bindings = False
        self.__build_python_bindings = False
        self.__ros_config = None
        self.__tcp_interface_config = None
        self.__local_path = None
        self.__allocator = RustAllocator.DefaultAllocator

    # ---------- GETTERS ---------------------------------------------

    @property
    def rebuild(self):
        """Whether to re-build the optimizer from scratch"""
        return self.__rebuild

    @property
    def id(self):
        """Unique identifier of build configuration"""
        return self.__id

    @property
    def cost_function_name(self):
        return self.__cost_function_name

    @property
    def grad_function_name(self):
        return self.__grad_cost_function_name

    @property
    def constraint_penalty_function_name(self):
        return self.__constraint_penalty_function

    @property
    def alm_mapping_f1_function_name(self):
        return self.__alm_constraints_mapping_f1

    @property
    def target_system(self):
        """Target system"""
        return self.__target_system

    @property
    def build_mode(self):
        """Build mode (release or debug)"""
        return self.__build_mode

    @property
    def build_dir(self):
        """Directory in which the auto-generated optimizer will be stored"""
        return self.__build_dir

    @property
    def open_version(self):
        """
        OpEn version used with the auto-generated solver

        :return: The method returns either a specific version of OpEn,
        which will be used with the auto-generated optimizer, or `None`,
        in which case, the latest version will be used. You may set your
        preferred version of OpEn with `with_open_version`
        """
        return self.__open_version

    @property
    def local_path(self):
        return self.__local_path

    @property
    def build_c_bindings(self):
        return self.__build_c_bindings

    @property
    def build_python_bindings(self):
        return self.__build_python_bindings

    @property
    def tcp_interface_config(self):
        return self.__tcp_interface_config

    @property
    def ros_config(self) -> RosConfiguration:
        """
        ROS package configuration
        :return: instance of RosConfiguration
        """
        return self.__ros_config

    @property
    def allocator(self) -> RustAllocator:
        """
        Memory allocator for generated Rust solver
        """
        return self.__allocator

    # ---------- SETTERS ---------------------------------------------

    def with_rebuild(self, do_rebuild):
        """
        Whether to clean and rebuild the code generator, if it already exists

        :param do_rebuild: if set to True, the target code generator
        will be cleaned and rebuilt from scratch

        :return: current instance of BuildConfiguration
        """
        self.__rebuild = do_rebuild
        return self

    def with_target_system(self, target_system):
        """
        Specify the target system

        :param target_system: target system as string (e.g., use
        "arm-unknown-linux-gnueabihf" or "rpi" for Raspberry Pi).
        Note that you must have installed the target using `rustup`
        if you need to cross-compile.

        :return: current instance of BuildConfiguration
        """
        if target_system.lower() == "rpi":
            self.__target_system = "arm-unknown-linux-gnueabihf"
        else:
            self.__target_system = target_system
        return self

    def with_build_mode(self, build_mode):
        """
        Set the build mode (debug/release)

        :param build_mode: Choose either 'debug' or 'release'; the former is
        fast, but suboptimal, while the later may take a while to compile,
        but the generated binary is significantly faster

        :return: current instance of BuildConfiguration

        """
        self.__build_mode = build_mode
        return self

    def with_build_directory(self, build_dir):
        """
        Specify the build directory

        :param build_dir: build directory as string

        :return: current instance of BuildConfiguration
        """
        self.__build_dir = build_dir
        return self

    def with_open_version(self, open_version="*", local_path=None):
        """
        Specify the version of OpEn to link to

        :param open_version: version of OpEn (in case you want to
        compile with an older version of OpEn; if not, the latest
        version of OpEn will be used)

        :param local_path: you can compile using a local version
        of OpEn. In that case, you need to provide the full absolute
        path to that local OpEn directory.

        :return: current instance of BuildConfiguration
        """
        self.__open_version = open_version
        self.__local_path = local_path
        return self

    def with_build_c_bindings(self, build_c_bindings=True):
        """
        If activated, OpEn will generate C/C++ bindings for the
        auto-generated solver

        :param build_c_bindings: whether to build C/C++ bindings for
        auto-generated solver; default: `True`, i.e., it suffices
        to call `build_config.with_build_c_bindings()` instead of
        `build_config.with_build_c_bindings(True)`

        :return: current instance of BuildConfiguration
        """
        self.__build_c_bindings = build_c_bindings
        return self

    def with_build_python_bindings(self, build_python_bindings=True):
        """
        If activated, OpEn will generate python bindings for the
        auto-generated solver

        :param build_python_bindings: whether to build python bindings for
        auto-generated solver; default: `True`, i.e., it suffices
        to call `build_config.with_build_python_bindings()` instead of
        `build_config.with_build_python_bindings(True)`

        :return: current instance of BuildConfiguration
        """
        self.__build_python_bindings = build_python_bindings
        return self

    def with_ros(self, ros_config: RosConfiguration):
        """
        Activates the generation of a ROS package. The caller must provide an
        instance of RosConfiguration
        :param ros_config: Configuation of ROS package
        :return: current instance of BuildConfiguration
        """
        self.__build_c_bindings = True  # no C++ bindings, no ROS package mate
        self.__ros_config = ros_config
        return self

    def with_tcp_interface_config(self, tcp_interface_config=TcpServerConfiguration()):
        """
        Specify a TCP server configuration object

        :param tcp_interface_config: Custom TCP server configuration

        :return: current instance of BuildConfiguration
        """
        self.__tcp_interface_config = tcp_interface_config
        return self

    def with_allocator(self, allocator: RustAllocator):
        """Specify a Rust memory allocator. 

        With this method the user can choose an alternative memory allocator
        such as `jemalloc` and `rpalloc`. For example, if you choose `jemalloc`
        as your memory allocator, the autogenerated project will have a 
        `Cargo.toml` file where `optimization-engine` is loaded as a dependency
        with the "jem" feature.

        :param allocator: allocator; instance of `RustAllocator`

        :return: current instance of BuildConfiguration
        """
        self.__allocator = allocator
        return self



class SolverStatus:
    """Class for storing the status of the solver."""
    def __init__(self, status):
        """Constructs instance of <code>SolverStatus</code>

        Args:
            status: dictionary containing solver status attributes

        Returns:
            New instance of <code>SolverStatus</code>
        """

        for k, v in status.items():
            attribute_name = "__{}".format(k)
            setattr(self, attribute_name, v)

    @property
    def exit_status(self):
        """
        Returns:
        Converged if the solver has converged

        NotConvergedIterations if the maximum number of outer or total inner
        iterations was reached

        NotConvergedOutOfTime if the solver did not have enough time to
        converge

        :return: The exit status of the solver
        """
        return self.__dict__["__exit_status"]

    @property
    def num_outer_iterations(self):
        """
        Returns the number of outer (ALM/PM) iterations of the algorithm
        :return: The number of outer iterations
        """
        return self.__dict__["__num_outer_iterations"]

    @property
    def num_inner_iterations(self):
        """
        Returns the total number of inner iterations for all inner problems
        :return: The total number of inner iterations
        """
        return self.__dict__["__num_inner_iterations"]

    @property
    def last_problem_norm_fpr(self):
        """
        Returns the infinity norm of the fixed-point residual of the last
        inner optimization problem
        :return: inf-norm of the FPR of the last inner problem
        """
        return self.__dict__["__last_problem_norm_fpr"]

    @property
    def f1_infeasibility(self):
        """
        Returns the distance between F1(u,p) and C at the solution, a measure
        of the infeasibility of the constraint F1(u,p) in C
        :return: Infeasibility of the constraint F1(u,p) in C
        """
        return self.__dict__["__delta_y_norm_over_c"]

    @property
    def f2_norm(self):
        """
        :return: Euclidean norm of f2(u,p) at the solution
        """
        return self.__dict__["__f2_norm"]

    @property
    def solve_time_ms(self):
        """
        :return: Total execution time in milliseconds
        """
        return self.__dict__["__solve_time_ms"]

    @property
    def penalty(self):
        """
        :return: Last value of the penalty parameter
        """
        return self.__dict__["__penalty"]

    @property
    def solution(self):
        """
        :return: Solution vector
        """
        return self.__dict__["__solution"]

    @property
    def lagrange_multipliers(self):
        """
        :return: Vector of Lagrange multipliers
        """
        return self.__dict__["__lagrange_multipliers"]

    @property
    def cost(self):
        """
        :return: Value of cost function at the solution
        """
        return self.__dict__["__cost"]



from .constraint import Constraint
import sys


class Halfspace(Constraint):
    """Halfspace

    A halfspace is a set of the form H = {c'x <= b}, where c is a given
    vector and b is a constant scalar.
    """

    def __init__(self, normal_vector, offset):
        """Construct a new halfspace H = {c'x <= b}

        :param normal_vector: vector c
        "param offset: parameter b
        """
        self.__normal_vector = [float(x) for x in normal_vector]
        self.__offset = offset

    @property
    def normal_vector(self):
        """
        Returns vector c
        :return:
        """
        return self.__normal_vector

    @property
    def offset(self):
        """
        Returns parameter b
        :return:
        """
        return self.__offset

    def dimension(self):
        """
        Dimension of the halfspace
        :return: length of normal vector
        """
        return len(self.__normal_vector)

    def project(self, u):
        raise NotImplementedError()

    def distance_squared(self, u):
        raise NotImplementedError()

    def is_convex(self):
        """
        A halfspace is a convex set
        :return:
        """
        return True

    def is_compact(self):
        """Whether the set is compact

        H is compact iff b < 0 and c = 0, in which case H is empty
        """
        eps = sys.float_info.epsilon
        # if b < 0 and c = 0, then H is empty, hence compact
        if self.offset < 0 and sum([self.normal_vector[idx]
                                    for idx in range(self.dimension())]) < eps:
            return True
        return False



import casadi.casadi as cs
import numpy as np
from .constraint import Constraint
import opengen.functions as fn


class Zero(Constraint):
    """A Euclidean ball constraint

    A constraint of the form ||u-u0|| <= r, where u0 is the center
    of the ball and r is its radius

    """

    def __init__(self):
        """
        Constructor for set Z = {0}

        """

    def distance_squared(self, u):
        return fn.norm2_squared(u)

    def project(self, u):
        raise NotImplementedError()

    def is_convex(self):
        return True

    def is_compact(self):
        return True



import casadi.casadi as cs
import numpy as np
from .constraint import Constraint
import opengen.functions as fn


class BallInf(Constraint):
    """Norm-ball of norm infinity translated by given vector

    Centered inf-ball around given point
    """

    def __init__(self, center=None, radius: float = 1.0):
        """Constructor for an infinity ball constraint

        Args:
            :param center: center of the ball; if this is equal to Null, the
                ball is centered at the origin

            :param radius: radius of the ball

        :return:
            New instance of Ballinf with given center and radius
        """

        if radius <= 0:
            raise Exception("The radius must be a positive number")

        if center is not None and not isinstance(center, (list, np.ndarray)):
            raise Exception("center is neither None nor a list nor np.ndarray")

        self.__center = None if center is None else np.array([float(i) for i in center])
        self.__radius = float(radius)

    @property
    def center(self):
        """Returns the center of the ball"""
        return self.__center

    @property
    def radius(self):
        """Returns the radius of the ball"""
        return self.__radius

    def norm_inf_fun_np(a):
        return np.linalg.norm(a, ord=np.inf)

    def distance_squared(self, u):
        # Function `distance` can be applied to CasADi symbols and
        # lists of numbers. However, if `u` is a symbol, we need to
        # use appropriate CasADi functions like cs.sign and cs.norm_2
        if fn.is_symbolic(u):
            # Case I: `u` is a CasADi SX symbol
            nu = u.size(1)
            v = u if self.__center is None else u - self.__center
        elif (isinstance(u, list) and all(isinstance(x, (int, float)) for x in u)) \
                or isinstance(u, np.ndarray):
            # Case II: `u` is an array of numbers or an np.ndarray
            nu = len(u)
            if self.__center is None:
                v = u
            else:
                # Note: self.__center is np.ndarray (`u` might be a list)
                z = self.__center.reshape(nu)
                u = np.array(u).reshape(nu)
                v = np.subtract(u, z)
        else:
            raise Exception("u is of invalid type")

        # Compute distance to Ball infinity:
        # dist^2(u) = norm(v)^2
        #            + SUM_i [
        #                         min{vi^2, r^2}
        #                       - 2*min{vi^2, r*|vi|}
        #                    ]
        # where v = u - xc

        squared_distance = fn.norm2(v)**2
        for i in range(nu):
            squared_distance += fn.fmin(v[i]**2, self.radius**2) \
                                - 2.0 * fn.fmin(v[i]**2, self.radius * fn.fabs(v[i]))
        return squared_distance

    def project(self, u):
        # Computes projection on infinity Ball
        raise NotImplementedError("Method `project` is not implemented")

    def is_convex(self):
        return True

    def is_compact(self):
        return True



class TcpServerConfiguration:

    def __init__(self, bind_ip='127.0.0.1', bind_port=8333):
        """
        Configuration of the TCP server

        :param bind_ip: IP address of generated TCP server. The default
        value is "127.0.0.1" (localhost). Use "0.0.0.0" for the generated
        TCP server to bind on all IPs.

        :param bind_port: Port on which the generated TCP server will bind.
        The default is 8333. Make sure you use an available port and avoid using
        privileged ports (i.e., 1 to 1024), well-known ports that are potentially
        used by other services and you should also avoid ephemeral ports
        (32768 to 65535 on Linux, 1025 to 5000 on Windows)

        :returns: new instance of TcpServerConfiguration, which can then be
        provided to an instance of `OpEnOptimizerBuilder` via `enable_tcp_interface`
        """
        self.__bind_ip = bind_ip
        self.__bind_port = bind_port

    @property
    def bind_ip(self):
        """
        IP at which the TCP server should bind, as a string
        :return: TCP server IP
        """
        return self.__bind_ip

    @property
    def bind_port(self):
        """
        Port at which the TCP server should bind, as int
        :return: TCP server port
        """
        return self.__bind_port



import casadi.casadi as cs
import numpy as np
from .constraint import Constraint
import opengen.functions as fn


class Ball2(Constraint):
    """A Euclidean ball constraint

    A constraint of the form ||u-u0|| <= r, where u0 is the center
    of the ball and r is its radius

    """

    def __init__(self, center=None, radius: float = 1.0):
        """Constructor for a Euclidean ball constraint

        Args:
            center: center of the ball; if this is equal to Null, the
            ball is centered at the origin

            radius: radius of the ball

        Returns:
            New instance of Ball2 with given center and radius
        """
        if radius <= 0:
            raise Exception("The radius must be a positive number")

        if center is not None and not isinstance(center, (list, np.ndarray)):
            raise Exception("center is neither None nor a list nor np.ndarray")

        self.__center = None if center is None else np.array(
            [float(i) for i in center])
        self.__radius = float(radius)

    @property
    def center(self):
        """Returns the center of the ball"""
        return self.__center

    @property
    def radius(self):
        """Returns the radius of the ball"""
        return self.__radius

    def distance_squared(self, u):
        """Computes the squared distance between a given point `u` and this ball

            :param u: given point; can be a list of float, a numpy
                n-dim array (`ndarray`) or a CasADi SX/MX symbol

            :return: distance from set as a float or a CasADi symbol
        """
        if fn.is_symbolic(u):
            # Case I: `u` is a CasADi SX symbol
            v = u if self.__center is None else u - self.__center
        elif (isinstance(u, list) and all(isinstance(x, (int, float)) for x in u))\
                or isinstance(u, np.ndarray):
            # Case II: `u` is an array of numbers or an np.ndarray
            if self.__center is None:
                v = u
            else:
                # Note: self.__center is np.ndarray (`u` might be a list)
                z = self.__center.reshape(len(u))
                u = np.array(u).reshape(len(u))
                v = np.subtract(u, z)
        else:
            raise Exception("u is of invalid type")

        # Compute squared distance
        # Let B = B(xc, r) be a Euclidean ball centered at xc with radius r
        #
        # dist_B^2(u) = max(0, sign(t(u))*t(u)^2), where
        # t(u) = ||u - x_c|| - r
        #
        # Note: there is another way to take squared distances:
        # d_B^2(u) = ||u||^2 * (1 - 1 / max{r, ||u||})^2,
        # but this leads to slightly lengthier CasADi symbols for
        # the Jacobian of the squared distance, so this approach was
        # abandoned
        t = fn.norm2(v) - self.radius
        return fn.fmax(0.0, fn.sign(t) * t ** 2)

    def project(self, u):
        # Idea: Computes projection on Ball as follows
        #   Proj_B(u) = u / max{1, ||u||},
        # which avoids dividing by zero or defining the projections
        raise NotImplementedError()

    def is_convex(self):
        return True

    def is_compact(self):
        return True



import casadi as cs


class SqDistSOC(cs.Callback):

    def __init__(self, name, opts={}):
        cs.Callback.__init__(self)
        self.construct(name, opts)

    def get_n_in(self):
        return 1

    def get_n_out(self):
        return 1

    def init(self):
        pass

    def eval(self, arg):
        x = arg[0]
        f = 0.5 * cs.dot(x, x)
        return [f]

    def has_jacobian(self, *_args):
        return False

    def has_forward(self, *args):
        return True

    def get_forward(self, *args):
        x = cs.MX.sym('x', 1)
        z = cs.MX.sym('x', 1)
        return cs.Function('fw', [x, z], [1])



from .solver_status import SolverStatus
from .solver_error import SolverError


class SolverResponse:
    """Stores a solver response of type SolverStatus or SolverError."""
    def __init__(self, d):
        """Constructs instance of <code>SolverResponse</code>

        Args:
            d: dictionary containing either status or error attributes

        Returns:
            New instance of <code>SolverResponse</code>
        """
        if 'Error' in d.values():
            self.__response = SolverError(d)
        else:
            self.__response = SolverStatus(d)

    def is_ok(self):
        """Determines if response is OK."""
        return isinstance(self.__response, SolverStatus)

    def get(self):
        """
        Returns response, which is an instance of SolverStatus,
        if the call was successful, or an instance of SolverError
        otherwise. It is recommended that you use is_ok() to check
        whether the call has succeeded first
        """
        return self.__response

    def __getitem__(self, key):
        return getattr(self.__response, key)



import casadi.casadi as cs
import numpy as np
from .is_numeric import is_numeric
from .is_symbolic import is_symbolic


def fabs(u):
    if is_numeric(u):
        return np.fabs(u)
    if is_symbolic(u):
        return cs.fabs(u)
    raise Exception("Illegal argument")



from .constraint import Constraint


class NoConstraints(Constraint):

    def __init__(self):
        pass

    def distance_squared(self, u):
        return 0.0

    def project(self, u):
        return u

    def is_convex(self):
        return True

    def is_compact(self):
        return False



import subprocess
import shutil
import datetime
import yaml
import warnings

import opengen.config as og_cfg
import opengen.definitions as og_dfn
import casadi.casadi as cs
import os
import jinja2
import logging
import pkg_resources
import sys

from .ros_builder import RosBuilder

_AUTOGEN_COST_FNAME = 'auto_casadi_cost.c'
_AUTOGEN_GRAD_FNAME = 'auto_casadi_grad.c'
_AUTOGEN_PNLT_CONSTRAINTS_FNAME = 'auto_casadi_mapping_f2.c'
_AUTOGEN_ALM_MAPPING_F1_FNAME = 'auto_casadi_mapping_f1.c'
_PYTHON_BINDINGS_PREFIX = 'python_bindings_'
_TCP_IFACE_PREFIX = 'tcp_iface_'
_ICASADI_PREFIX = 'icasadi_'
_ROS_PREFIX = 'ros_node_'


def make_dir_if_not_exists(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


class OpEnOptimizerBuilder:
    """Builder for code generation

    This class is a builder that can be used to generate
    code for a parametric optimizer.
    """

    def __init__(self,
                 problem,
                 metadata=og_cfg.OptimizerMeta(),
                 build_configuration=og_cfg.BuildConfiguration(),
                 solver_configuration=og_cfg.SolverConfiguration()):
        """Constructor of OpEnOptimizerBuilder

        Args:
            problem: instance of \link opengen.builder.problem.Problem Problem \endlink
            metadata: instance of \link opengen.config.meta.OptimizerMeta OptimizerMeta \endlink
            build_configuration: instance of \link opengen.config.build_config.BuildConfiguration BuildConfiguration \endlink
            solver_configuration: instance of \link  opengen.config.solver_config.SolverConfiguration SolverConfiguration \endlink

        Returns:
            New instance of OpEnOptimizerBuilder.
        """
        self.__problem = problem
        self.__meta = metadata
        self.__build_config = build_configuration
        self.__solver_config = solver_configuration
        self.__generate_not_build = False
        self.__logger = logging.getLogger(
            'opengen.builder.OpEnOptimizerBuilder')
        self.with_verbosity_level(1)

    @staticmethod
    def __get_template(name, subdir=None):
        file_loader = jinja2.FileSystemLoader(og_dfn.templates_subdir(subdir))
        env = jinja2.Environment(loader=file_loader, autoescape=True)
        return env.get_template(name)

    def with_verbosity_level(self, verbosity_level):
        """Specify the verbosity level

        :param verbosity_level: level of verbosity (0,1,2,3)

        :returns: Current builder object
        """
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(verbosity_level)
        c_format = logging.Formatter('[%(levelname)s] %(message)s')
        stream_handler.setFormatter(c_format)
        self.__logger.setLevel(verbosity_level)
        self.__logger.handlers.clear()
        self.__logger.addHandler(stream_handler)

        return self

    def with_problem(self, problem):
        """Specify problem

        :param problem: optimization problem data

        :returns: Current builder object
        """
        self.__problem = problem
        return self

    def with_generate_not_build_flag(self, flag):
        """Whether to build (or just generate code)

        If set to true, the code will be generated, but it will not be
        build (mainly for debugging purposes)

        :param flag: generate and not build

        :returns: Current builder object
        """
        self.__generate_not_build = flag
        return self

    def __make_build_command(self):
        """
        Cargo build command (possibly, with --release)

        """
        command = ['cargo', 'build', '-q']
        if self.__build_config.build_mode.lower() == 'release':
            command.append('--release')

        if self.__build_config.target_system is not None:
            command.append('--target='+self.__build_config.target_system)

        return command

    def __target_dir(self):
        """

        Target directory

        """
        return os.path.abspath(
            os.path.join(
                self.__build_config.build_dir,
                self.__meta.optimizer_name))

    def __icasadi_target_dir(self):
        """

        Returns icasadi target directory (instance of os.path)

        """
        icasadi_target_dir_name = _ICASADI_PREFIX + self.__meta.optimizer_name
        return os.path.abspath(
            os.path.join(
                self.__build_config.build_dir,
                self.__meta.optimizer_name, icasadi_target_dir_name))

    def __ros_target_dir(self):
        ros_target_dir_name = _ROS_PREFIX + self.__meta.optimizer_name
        return os.path.abspath(
            os.path.join(
                self.__build_config.build_dir,
                self.__meta.optimizer_name, ros_target_dir_name))

    def __prepare_target_project(self):
        """Creates folder structure

        Creates necessary folders
        Runs `cargo init` in that folder

        """
        self.__logger.info("Creating necessary folders")

        # Create target directory if it does not exist
        target_dir = self.__target_dir()
        if self.__build_config.rebuild:
            if os.path.exists(target_dir) and os.path.isdir(target_dir):
                shutil.rmtree(target_dir)
            os.makedirs(target_dir)
        else:
            make_dir_if_not_exists(target_dir)

    def __copy_icasadi_to_target(self):
        """

        Copy 'icasadi' folder and its contents into target directory

        """
        self.__logger.info("Copying icasadi interface to target directory")
        origin_icasadi_dir = og_dfn.original_icasadi_dir()
        target_icasadi_dir = self.__icasadi_target_dir()
        if not os.path.exists(target_icasadi_dir):
            os.makedirs(target_icasadi_dir)
        shutil.rmtree(target_icasadi_dir)
        shutil.copytree(origin_icasadi_dir,
                        target_icasadi_dir,
                        ignore=shutil.ignore_patterns(
                            '*.lock', 'ci*', 'target', 'auto*'))

    def __generate_icasadi_cargo_toml(self):
        """
        Generate icasadi's Cargo.toml file
        """
        self.__logger.info("Generating icasadi's Cargo.toml")
        icasadi_cargo_template = OpEnOptimizerBuilder.__get_template(
            'icasadi_cargo.toml', subdir='icasadi')
        icasadi_cargo_output_template = icasadi_cargo_template.render(
            meta=self.__meta)
        icasadi_cargo_allocator_path = os.path.abspath(
            os.path.join(self.__icasadi_target_dir(), "Cargo.toml"))
        with open(icasadi_cargo_allocator_path, "w") as fh:
            fh.write(icasadi_cargo_output_template)

    def __generate_icasadi_c_interface(self):
        """
        Generates the C interface file interface.c

        """
        self.__logger.info("Generating intercafe.c (C interface)")
        cint_template = OpEnOptimizerBuilder.__get_template(
            'interface.c', 'icasadi')
        cint_output_template = cint_template.render(
            meta=self.__meta,
            problem=self.__problem,
            build_config=self.__build_config,
            timestamp_created=datetime.datetime.now())
        cint_icallocator_path = os.path.abspath(
            os.path.join(self.__icasadi_target_dir(), "extern", "interface.c"))
        with open(cint_icallocator_path, "w") as fh:
            fh.write(cint_output_template)

    def __generate_icasadi_lib(self):
        """
        Generates the Rust library file of icasadi

        Generates src/lib.rs
        """
        self.__logger.info("Generating icasadi Rust library file")
        icasadi_lib_template = OpEnOptimizerBuilder.__get_template(
            'icasadi_lib.rs', 'icasadi')
        icasadi_lib_output_template = icasadi_lib_template.render(meta=self.__meta,
                                                                  problem=self.__problem,
                                                                  build_config=self.__build_config,
                                                                  timestamp_created=datetime.datetime.now())
        icasadi_lib_rs_path = os.path.abspath(
            os.path.join(self.__icasadi_target_dir(), "src", "lib.rs"))
        with open(icasadi_lib_rs_path, "w") as fh:
            fh.write(icasadi_lib_output_template)

    def __generate_cargo_toml(self):
        """
        Generates Cargo.toml for generated project

        """
        self.__logger.info("Generating Cargo.toml for target optimizer")
        target_dir = self.__target_dir()
        cargo_template = OpEnOptimizerBuilder.__get_template(
            'optimizer_cargo.toml')
        cargo_output_template = cargo_template.render(
            meta=self.__meta,
            build_config=self.__build_config,
            activate_tcp_server=self.__build_config.tcp_interface_config is not None,
            activate_clib_generation=self.__build_config.build_c_bindings,
            timestamp_created=datetime.datetime.now())
        cargo_toml_path = os.path.abspath(
            os.path.join(target_dir, "Cargo.toml"))
        with open(cargo_toml_path, "w") as fh:
            fh.write(cargo_output_template)

    def __generate_memory_code(self, cost=None, grad=None, f1=None, f2=None):
        """
        Creates file casadi_memory.h with memory sizes

        :param cost: cost function (cs.Function)
        :param grad: grad function (cs.Function)
        :param f1: mapping F1  (cs.Function)
        :param f2: mapping F2  (cs.Function)
        """
        self.__logger.info("Generating casadi_memory.h")
        casadi_mem_template = OpEnOptimizerBuilder.__get_template(
            'casadi_memory.h', 'icasadi')
        casadi_mem_output_template = casadi_mem_template.render(
            cost=cost, grad=grad,
            f1=f1, f2=f2,
            build_config=self.__build_config,
            meta=self.__meta,
            timestamp_created=datetime.datetime.now())
        memory_path = os.path.abspath(
            os.path.join(self.__icasadi_target_dir(), "extern", "casadi_memory.h"))
        with open(memory_path, "w") as fh:
            fh.write(casadi_mem_output_template)

    def __construct_function_psi(self):
        """
        Construct function psi and its gradient

        :return: cs.Function objects: psi_fun, grad_psi_fun
        """
        self.__logger.info("Defining function psi(u, xi, p) and its gradient")
        problem = self.__problem
        bconfig = self.__build_config
        u = problem.decision_variables
        p = problem.parameter_variables
        n2 = problem.dim_constraints_penalty()
        n1 = problem.dim_constraints_aug_lagrangian()
        phi = problem.cost_function
        alm_set_c = problem.alm_set_c
        f1 = problem.penalty_mapping_f1
        f2 = problem.penalty_mapping_f2

        psi = phi

        if n1 + n2 > 0:
            n_xi = n1 + 1
        else:
            n_xi = 0

        xi = cs.SX.sym('xi', n_xi, 1) if isinstance(u, cs.SX) \
            else cs.MX.sym('xi', n_xi, 1)

        # Note: In the first term below, we divide by 'max(c, 1)', instead of
        #       just 'c'. The reason is that this allows to set c=0 and
        #       retrieve the value of the original cost function
        if n1 > 0:
            sq_dist_term = alm_set_c.distance_squared(
                f1 + xi[1:n1+1]/cs.fmax(xi[0], 1))
            psi += xi[0] * sq_dist_term / 2

        if n2 > 0:
            psi += xi[0] * cs.dot(f2, f2) / 2

        jac_psi = cs.gradient(psi, u)

        psi_fun = cs.Function(bconfig.cost_function_name, [u, xi, p], [psi])
        grad_psi_fun = cs.Function(
            bconfig.grad_function_name, [u, xi, p], [jac_psi])
        return psi_fun, grad_psi_fun

    def __construct_mapping_f1_function(self) -> cs.Function:
        self.__logger.info("Defining function F1(u, p)")
        problem = self.__problem
        u = problem.decision_variables
        p = problem.parameter_variables
        n1 = problem.dim_constraints_aug_lagrangian()
        f1 = problem.penalty_mapping_f1

        if n1 > 0:
            mapping_f1 = f1
        else:
            mapping_f1 = 0

        alm_mapping_f1_fun = cs.Function(
            self.__build_config.alm_mapping_f1_function_name,
            [u, p], [mapping_f1])
        return alm_mapping_f1_fun

    def __construct_mapping_f2_function(self) -> cs.Function:
        self.__logger.info("Defining function F2(u, p)")
        problem = self.__problem
        u = problem.decision_variables
        p = problem.parameter_variables
        n2 = problem.dim_constraints_penalty()

        if n2 > 0:
            penalty_constraints = problem.penalty_mapping_f2
        else:
            penalty_constraints = 0

        alm_mapping_f2_fun = cs.Function(
            self.__build_config.constraint_penalty_function_name,
            [u, p], [penalty_constraints])

        return alm_mapping_f2_fun

    def __generate_casadi_code(self):
        """Generates CasADi C code"""
        self.__logger.info("Defining CasADi functions and generating C code")
        bconfig = self.__build_config

        # -----------------------------------------------------------------------
        psi_fun, grad_psi_fun = self.__construct_function_psi()
        cost_file_name = bconfig.cost_function_name + ".c"
        grad_file_name = bconfig.grad_function_name + ".c"
        self.__logger.info("Function psi and its gradient (C code)")
        psi_fun.generate(cost_file_name)
        grad_psi_fun.generate(grad_file_name)
        icasadi_extern_dir = os.path.join(
            self.__icasadi_target_dir(), "extern")
        shutil.move(cost_file_name, os.path.join(
            icasadi_extern_dir, _AUTOGEN_COST_FNAME))
        shutil.move(grad_file_name, os.path.join(
            icasadi_extern_dir, _AUTOGEN_GRAD_FNAME))

        # -----------------------------------------------------------------------
        mapping_f1_fun = self.__construct_mapping_f1_function()
        f1_file_name = bconfig.alm_mapping_f1_function_name + ".c"
        self.__logger.info("Mapping F1 (C code)")
        mapping_f1_fun.generate(f1_file_name)
        # Move auto-generated file to target folder
        shutil.move(f1_file_name,
                    os.path.join(icasadi_extern_dir, _AUTOGEN_ALM_MAPPING_F1_FNAME))

        # -----------------------------------------------------------------------
        mapping_f2_fun = self.__construct_mapping_f2_function()
        f2_file_name = bconfig.constraint_penalty_function_name + ".c"
        self.__logger.info("Mapping F2 (C code)")
        mapping_f2_fun.generate(f2_file_name)
        # Move auto-generated file to target folder
        shutil.move(f2_file_name,
                    os.path.join(icasadi_extern_dir, _AUTOGEN_PNLT_CONSTRAINTS_FNAME))

        self.__generate_memory_code(psi_fun, grad_psi_fun,
                                    mapping_f1_fun, mapping_f2_fun)

    def __generate_main_project_code(self):
        self.__logger.info(
            "Generating main code for target optimizer (lib.rs)")
        target_dir = self.__target_dir()
        optimizer_rs_template = OpEnOptimizerBuilder.__get_template(
            'optimizer.rs')
        optimizer_rs_output_template = optimizer_rs_template.render(
            solver_config=self.__solver_config,
            meta=self.__meta,
            problem=self.__problem,
            timestamp_created=datetime.datetime.now(),
            activate_clib_generation=self.__build_config.build_c_bindings,
            float=float)
        target_source_path = os.path.join(target_dir, "src")
        target_scr_lib_rs_path = os.path.join(target_source_path, "lib.rs")
        make_dir_if_not_exists(target_source_path)
        with open(target_scr_lib_rs_path, "w") as fh:
            fh.write(optimizer_rs_output_template)

    def __generate_build_rs(self):
        self.__logger.info("Generating build.rs for target optimizer")
        target_dir = self.__target_dir()
        build_rs_template = OpEnOptimizerBuilder.__get_template(
            'optimizer_build.rs')
        build_rs_output_template = build_rs_template.render(
            meta=self.__meta,
            activate_clib_generation=self.__build_config.build_c_bindings)
        target_build_lib_rs_path = os.path.join(target_dir, "build.rs")
        with open(target_build_lib_rs_path, "w") as fh:
            fh.write(build_rs_output_template)

    def __build_optimizer(self):
        target_dir = os.path.abspath(self.__target_dir())
        command = self.__make_build_command()
        p = subprocess.Popen(command, cwd=target_dir)
        process_completion = p.wait()
        if process_completion != 0:
            raise Exception('Rust build failed')

    def __build_python_bindings(self):
        self.__logger.info("Building Python bindings")
        target_dir = os.path.abspath(self.__target_dir())
        optimizer_name = self.__meta.optimizer_name
        python_bindings_dir_name = _PYTHON_BINDINGS_PREFIX + optimizer_name
        python_bindings_dir = os.path.join(
            target_dir, python_bindings_dir_name)
        command = self.__make_build_command()
        p = subprocess.Popen(command, cwd=python_bindings_dir)
        process_completion = p.wait()
        if process_completion != 0:
            raise Exception('Rust build of Python bindings failed')

        if self.__build_config.build_mode.lower() == 'release':
            build_dir = os.path.join(python_bindings_dir, 'target', 'release')
        else:
            build_dir = os.path.join(python_bindings_dir, 'target', 'debug')

        pltform_extension_dict = {'linux': ('.so', '.so'),
                                  'darwin': ('.dylib', '.so'),
                                  'win32': ('.dll', '.pyd')}
        (original_lib_extension,
         target_lib_extension) = pltform_extension_dict[sys.platform]
        optimizer_prefix = "lib" if sys.platform != "win32" else ""
        generated_bindings = os.path.join(
            build_dir, "{}{}{}".format(optimizer_prefix, optimizer_name, original_lib_extension))
        target_bindings = os.path.join(
            target_dir, "{}{}".format(optimizer_name, target_lib_extension))
        shutil.copyfile(generated_bindings, target_bindings)
        self.__logger.info("To use the Python bindings do:\n\n"
                           "       import sys\n"
                           "       sys.path.insert(1, \"{}\")\n"
                           "       import {}\n"
                           "       solver = {}.solver()".format(target_dir, optimizer_name, optimizer_name))

    def __build_tcp_iface(self):
        self.__logger.info("Building the TCP interface")
        target_dir = os.path.abspath(self.__target_dir())
        optimizer_name = self.__meta.optimizer_name
        tcp_iface_dir_name = _TCP_IFACE_PREFIX + optimizer_name
        tcp_iface_dir = os.path.join(target_dir, tcp_iface_dir_name)
        command = self.__make_build_command()
        p = subprocess.Popen(command, cwd=tcp_iface_dir)
        process_completion = p.wait()
        if process_completion != 0:
            raise Exception('Rust build of TCP interface failed')

    def __initialize(self):
        self.__logger.info("--- Initialising builder: '%s'" %
                           self.__meta.optimizer_name)

    def __check_user_provided_parameters(self):
        self.__logger.info("Checking user parameters")

    def __generate_code_python_bindings(self):
        self.__logger.info("Generating code for Python bindings")
        target_dir = self.__target_dir()
        python_bindings_dir_name = _PYTHON_BINDINGS_PREFIX + self.__meta.optimizer_name
        python_bindings_dir = os.path.join(
            target_dir, python_bindings_dir_name)
        python_bindings_source_dir = os.path.join(python_bindings_dir, "src")
        self.__logger.info(
            "Generating code for Python bindings in {}".format(python_bindings_dir))

        # make python_bindings/ and python_bindings/src
        make_dir_if_not_exists(python_bindings_dir)
        make_dir_if_not_exists(python_bindings_source_dir)

        # generate tcp_server.rs for python_bindings
        python_rs_template = OpEnOptimizerBuilder.__get_template(
            'python_bindings.rs', 'python')
        python_rs_output_template = python_rs_template.render(
            meta=self.__meta,
            timestamp_created=datetime.datetime.now())
        target_python_rs_path = os.path.join(
            python_bindings_source_dir, "lib.rs")
        with open(target_python_rs_path, "w") as fh:
            fh.write(python_rs_output_template)

        # generate Cargo.toml for python_bindings
        python_rs_template = OpEnOptimizerBuilder.__get_template(
            'python_bindings_cargo.toml', 'python')
        python_rs_output_template = python_rs_template.render(
            meta=self.__meta,
            build_config=self.__build_config,
            timestamp_created=datetime.datetime.now())
        target_python_rs_path = os.path.join(python_bindings_dir, "Cargo.toml")
        with open(target_python_rs_path, "w") as fh:
            fh.write(python_rs_output_template)

        # move cargo_config into .cargo/config
        target_cargo_config_dir = os.path.join(python_bindings_dir, '.cargo')
        make_dir_if_not_exists(target_cargo_config_dir)
        cargo_config_file = os.path.join(
            og_dfn.templates_dir(), 'python', 'cargo_config')
        shutil.copy(cargo_config_file, os.path.join(
            target_cargo_config_dir, 'config'))

    def __generate_code_tcp_interface(self):
        self.__logger.info(
            "Generating code for TCP/IP interface (tcp_iface/src/main.rs)")
        self.__logger.info("TCP server will bind at %s:%d",
                           self.__build_config.tcp_interface_config.bind_ip,
                           self.__build_config.tcp_interface_config.bind_port)
        target_dir = self.__target_dir()
        tcp_iface_dir_name = _TCP_IFACE_PREFIX + self.__meta.optimizer_name
        tcp_iface_dir = os.path.join(target_dir, tcp_iface_dir_name)
        tcp_iface_source_dir = os.path.join(tcp_iface_dir, "src")

        # make tcp_iface/ and tcp_iface/src
        make_dir_if_not_exists(tcp_iface_dir)
        make_dir_if_not_exists(tcp_iface_source_dir)

        # generate tcp_server.rs for tcp_iface
        tcp_rs_template = OpEnOptimizerBuilder.__get_template(
            'tcp_server.rs', 'tcp')
        tcp_rs_output_template = tcp_rs_template.render(
            meta=self.__meta,
            tcp_server_config=self.__build_config.tcp_interface_config,
            timestamp_created=datetime.datetime.now())
        target_tcp_rs_path = os.path.join(tcp_iface_source_dir, "main.rs")
        with open(target_tcp_rs_path, "w") as fh:
            fh.write(tcp_rs_output_template)

        # generate Cargo.toml for tcp_iface
        tcp_rs_template = OpEnOptimizerBuilder.__get_template(
            'tcp_server_cargo.toml', 'tcp')
        tcp_rs_output_template = tcp_rs_template.render(
            meta=self.__meta,
            build_config=self.__build_config,
            timestamp_created=datetime.datetime.now())
        target_tcp_rs_path = os.path.join(tcp_iface_dir, "Cargo.toml")
        with open(target_tcp_rs_path, "w") as fh:
            fh.write(tcp_rs_output_template)

    def __generate_yaml_data_file(self):
        self.__logger.info("Generating YAML configuration file")
        tcp_config = self.__build_config.tcp_interface_config
        metadata = self.__meta
        build_config = self.__build_config
        solver_config = self.__solver_config
        target_dir = self.__target_dir()

        target_yaml_file_path = os.path.join(target_dir, "optimizer.yml")

        opengen_version = pkg_resources.require("opengen")[0].version

        tcp_details = None if tcp_config is None \
            else {'ip': tcp_config.bind_ip, 'port': tcp_config.bind_port}
        metadata_details = {'optimizer_name': metadata.optimizer_name,
                            'version': metadata.version,
                            'authors': metadata.authors,
                            'licence': metadata.licence}
        build_details = {'id': build_config.id,
                         'open_version': build_config.open_version,
                         'opengen_version': opengen_version,
                         'build_dir': build_config.build_dir,
                         'build_mode': build_config.build_mode,
                         'target_system': build_config.target_system,
                         'cost_function_name': build_config.cost_function_name,
                         'grad_function_name': build_config.grad_function_name,
                         'mapping_f2': build_config.constraint_penalty_function_name,
                         'mapping_f1': build_config.alm_mapping_f1_function_name
                         }
        solver_details = {'lbfgs_memory': solver_config.lbfgs_memory,
                          'tolerance': solver_config.tolerance,
                          'constraints_tolerance': solver_config.constraints_tolerance,
                          'penalty_weight_update_factor': solver_config.penalty_weight_update_factor,
                          'max_outer_iterations': solver_config.max_outer_iterations,
                          'max_inner_iterations': solver_config.max_inner_iterations,
                          'max_duration_micros': solver_config.max_duration_micros
                          }
        details = {'meta': metadata_details, 'tcp': tcp_details, 'build': build_details,
                   'solver': solver_details,
                   "_comment": "auto-generated file; do not modify"}
        with open(target_yaml_file_path, 'w') as outfile:
            yaml.dump(details, outfile, Dumper=yaml.Dumper)

    def enable_tcp_interface(self,
                             tcp_server_configuration=og_cfg.TcpServerConfiguration()):
        # This method should not be used!
        raise DeprecationWarning(
            "deprecated (use BuildConfiguration.with_tcp_interface_config instead)")

    def __generate_c_bindings_example(self):
        self.__logger.info(
            "Generating example_optimizer.c (C bindings example for your convenience)")
        target_dir = self.__target_dir()
        cbind_template = OpEnOptimizerBuilder.__get_template(
            'example_optimizer_c_bindings.c', 'c')
        cbind_output_template = cbind_template.render(
            meta=self.__meta,
            build_config=self.__build_config,
            problem=self.__problem)
        cbind_target_path = os.path.join(target_dir, "example_optimizer.c")
        with open(cbind_target_path, "w") as fh:
            fh.write(cbind_output_template)

    def __generate_c_bindings_makefile(self):
        self.__logger.info(
            "Generating CMakeLists (do: `cmake .; make` to compile the autogenerated example)")
        target_dir = self.__target_dir()
        cbind_makefile_template = OpEnOptimizerBuilder.__get_template(
            'example_cmakelists.txt', 'c')
        cbind_makefile_output_template \
            = cbind_makefile_template.render(meta=self.__meta,
                                             build_config=self.__build_config)
        cbind_makefile_target_path = os.path.join(target_dir, "CMakeLists.txt")
        with open(cbind_makefile_target_path, "w") as fh:
            fh.write(cbind_makefile_output_template)

    def build(self):
        """Generate code and build project

        Raises:
            Exception: if the build process fails
            Exception: if there some parameters have wrong, inadmissible or incompatible values

        """
        self.__initialize()                      # initialize default value (if not provided)
        self.__check_user_provided_parameters()  # check the provided parameters
        self.__prepare_target_project()          # create folders; init cargo project
        self.__copy_icasadi_to_target()          # copy icasadi/ files to target dir
        self.__generate_icasadi_cargo_toml()     # generate icasadi's Cargo.toml file
        self.__generate_cargo_toml()             # generate Cargo.toml using template
        self.__generate_icasadi_lib()            # generate icasadi lib.rs
        # generate all necessary CasADi C files:
        self.__generate_casadi_code()
        #                                        #   - auto_casadi_cost.c
        #                                        #   - auto_casadi_grad.c
        #                                        #   - auto_casadi_mapping_f1.c
        #                                        #   - auto_casadi_mapping_f2.c
        #                                        #   - casadi_memory.h
        self.__generate_icasadi_c_interface()    # generate icasadi/extern/interface.c
        # generate main part of code (at build/{name}/src/main.rs)
        self.__generate_main_project_code()
        self.__generate_build_rs()               # generate build.rs file
        self.__generate_yaml_data_file()         # create YAML file with metadata

        if not self.__generate_not_build:
            self.__logger.info("Building optimizer")
            self.__build_optimizer()             # build overall project

        if self.__build_config.tcp_interface_config is not None:
            self.__logger.info("Generating TCP/IP server")
            self.__generate_code_tcp_interface()
            if not self.__generate_not_build:
                self.__build_tcp_iface()

        if self.__build_config.build_c_bindings:
            self.__logger.info("Generating C/C++ bindings")
            self.__generate_c_bindings_example()
            self.__generate_c_bindings_makefile()

        if self.__build_config.build_python_bindings:
            self.__logger.info("Generating Python bindings")
            self.__generate_code_python_bindings()
            if not self.__generate_not_build:
                self.__build_python_bindings()

        if self.__build_config.ros_config is not None:
            ros_builder = RosBuilder(
                self.__meta,
                self.__build_config,
                self.__solver_config)
            ros_builder.build()



from .constraint import Constraint
import opengen.functions as fn


class Rectangle(Constraint):
    """A Rectangle (Box) constraint"""

    @classmethod
    def __check_xmin_xmax(cls, xmin, xmax):
        # (None, None) is not allowed
        if xmin is None and xmax is None:
            raise Exception("At least one of xmin and xmax must be not None")

        # If one is not None, it must be a list
        if xmin is not None and not isinstance(xmin, list):
            raise Exception("xmin is neither None nor a list")

        if xmax is not None and not isinstance(xmax, list):
            raise Exception("xmax is neither None nor a list")

        # If BOTH are not None, they must be compatible
        if xmin is not None and xmax is not None:
            if len(xmin) != len(xmax):
                raise Exception("xmin and xmax must have equal lengths")
            for (xmin_element, xmax_element) in zip(xmin, xmax):
                if xmin_element > xmax_element:
                    raise Exception("xmin must be <= xmax")

    def __init__(self, xmin, xmax):
        """Construct a new instance of Rectangle

        Args:
            xmin: minimum bounds (can be None)
            xmax: maximum bounds (can be None)

        Raises:
            Exception: if both xmin and xmax is None
              Exception: if xmin/xmax is not None and not a list (wrong type)
            Exception: if xmin and xmax have incompatible lengths
            Exception: if xmin(i) > xmax(i) for some i (empty set)

        Returns:
             A new instance of Rectangle
        """
        Rectangle.__check_xmin_xmax(xmin, xmax)

        # Store xmin and xmax in attributes
        self.__xmin = None if xmin is None else [float(i) for i in xmin]
        self.__xmax = None if xmax is None else [float(i) for i in xmax]

    @property
    def xmin(self):
        """Minimum bound"""
        return self.__xmin

    @property
    def xmax(self):
        """Maximum bound"""
        return self.__xmax

    def dimension(self):
        if self.__xmin is not None:
            return len(self.__xmin)
        if self.__xmax is not None:
            return len(self.__xmax)
        raise Exception("Absurd: both xmin and xmax are None!")

    def idx_bound_finite_all(self):
        idx_both_finite = []

        if self.__xmin is None or self.__xmax is None:
            return idx_both_finite

        for i in range(self.dimension()):
            xmini = self.__xmin[i]
            xmaxi = self.__xmax[i]
            if xmini > float('-inf') and xmaxi < float('inf'):
                idx_both_finite += [i]

        return idx_both_finite

    def idx_infinite_only_xmin(self):
        idx_xmin_infinite = []

        if self.__xmax is None:
            # If xmax is None (infinite), we should return
            # the empty set
            return idx_xmin_infinite

        # Hereafter, xmax is not None (but xmin can be None)
        for i in range(self.dimension()):
            xmini = self.__xmin[i] if self.__xmin is not None else float('-inf')
            xmaxi = self.__xmax[i]
            if xmini == float('-inf') and xmaxi < float('inf'):
                idx_xmin_infinite += [i]

        return idx_xmin_infinite

    def idx_infinite_only_xmax(self):
        idx_xmin_infinite = []

        if self.__xmin is None:
            # If xmin is None (-infinity), we should return
            # the empty set
            return idx_xmin_infinite

        # Hereafter, xmin is not None (xmax might be)
        for i in range(self.dimension()):
            xmini = self.__xmin[i]
            xmaxi = self.__xmax[i] if self.__xmax is not None else float('inf')
            if xmaxi == float('inf') and xmini > float('-inf'):
                idx_xmin_infinite += [i]

        return idx_xmin_infinite

    def distance_squared(self, u):
        idx1 = self.idx_infinite_only_xmin()
        idx2 = self.idx_infinite_only_xmax()
        idx3 = self.idx_bound_finite_all()

        dist_sq = 0.0
        for i in idx1:
            dist_sq += fn.fmax(0.0, u[i] - self.__xmax[i]) ** 2

        for i in idx2:
            dist_sq += fn.fmin(0.0, u[i] - self.__xmin[i]) ** 2

        for i in idx3:
            dist_sq += fn.fmin(
                fn.fmax(0.0, u[i] - self.__xmax[i]),
                u[i] - self.__xmin[i]) ** 2

        return dist_sq

    def project(self, u):
        raise NotImplementedError()

    def is_convex(self):
        return True

    def is_compact(self):
        if self.__xmin is None:
            return False

        if self.__xmax is None:
            return False

        for i in range(len(self.__xmin)):
            if self.__xmin[i] == float('-inf'):
                return False

        for i in range(len(self.__xmax)):
            if self.__xmax[i] == float('inf'):
                return False

        return True


import casadi.casadi as cs
import numpy as np
from .constraint import Constraint
import opengen.functions as fn


class FiniteSet(Constraint):
    """Finite sets
    """

    def __init__(self, points=None):
        """Constructor for a finite set
        """
        if points is not None and len(points) > 0:
            first_point_len = len(points[0])
            for point in points[1:]:
                point_len = len(point)
                if point_len != first_point_len:
                    raise Exception("Invalid input (points have unequal dimensions)")
        self.__points = None if points is None else [[float(x) for x in p] for p in points]

    @property
    def points(self):
        return self.__points

    def dimension(self):
        p = self.points
        if p is None or len(p) == 0:
            return 0
        return len(p[0])

    def cardinality(self):
        p = self.points
        if p is None:
            return 0
        return len(p)

    def distance_squared(self, u):
        raise NotImplementedError()

    def project(self, u):
        raise NotImplementedError()

    def is_convex(self):
        return self.cardinality() == 1 and self.dimension() > 0

    def is_compact(self):
        return True



import casadi.casadi as cs
from .set_y_calculator import SetYCalculator
from ..constraints.no_constraints import NoConstraints


class Problem:
    """Definition of an optimization problem

    Provides the cost function, constraints and additional
    ALM/penalty-type constraints.
    """

    def __init__(self, u, p, cost):
        """Construct an optimization problem

            :param u: decision variable (CasADi variable)
            :param p: parameter (CasADi variable)
            :param cost: cost function (CasADi function of u and p)

        Example:
            >>> import casadi.casadi as cs
            >>> import opengen as og
            >>> # Define u and p
            >>> u = cs.SX.sym('u', 5)
            >>> p = cs.SX.sym('p', 2)
            >>> # Cost function
            >>> phi = og.functions.rosenbrock(u, p)
            >>> # Define optimization problem
            >>> problem = og.builder.Problem(u, p, phi)

        """

        # Decision variable: u
        self.__u = u
        # Parameter: p
        self.__p = p
        # Cost function: f(u, p)
        self.__cost = cost
        # Constraints on u: u in U
        self.__u_constraints = NoConstraints()
        # ALM-type constraints: mapping F1(u, p)
        self.__alm_mapping_f1 = None
        # ALM-type constraints: set C
        # Corresponds to the constraints: F1(u, p) in C
        self.__alm_set_c = None
        # Set Y of dual variables (compact subset of C*)
        self.__alm_set_y = None
        # Penalty-type constraints: mapping F2(u, p)
        # Constraints: F1(u, p) = 0
        self.__penalty_mapping_f2 = None
        # penalty function - default (and most commonly used): squared 2-norm
        self.__penalty_function = None

    # ---------- SETTERS -----------------------------------------------

    def with_constraints(self, u_constraints):
        """Specify or update the constraints of the problem

        Args:
            u_constraints: constraints on the decision variable; must
                           be a Constraint object (such as
                           opengen.constraints.ball2.Ball2
                           and opengen.constraints.rectangle.Rectangle)

        Returns:
            Current object

        """
        self.__u_constraints = u_constraints
        return self

    def with_penalty_constraints(self, penalty_constraints):
        """Constraints to for the penalty method

        Specify the constraints to be treated with the penalty method (that is,
        function F2(u; p))

        :param penalty_constraints: a function <code>c(u, p)</code>, of the decision
                variable <code>u</code> and the parameter vector <code>p</code>, which
                corresponds to the constraints <code>c(u, p)</code>

        :return: self
        """

        self.__penalty_mapping_f2 = penalty_constraints
        return self

    def with_aug_lagrangian_constraints(self, mapping_f1, set_c, set_y=None):
        """
        Constraints: F1(u, p) in C

        :param mapping_f1: mapping of the form `F1: R^{n} x R^{p} --> R^{n1}`
        :param set_c: a convex closed set C
        :param set_y: a compact subset of C*, the convex conjugate of C

        :return: self
        """
        if not set_c.is_convex():
            raise Exception("Set C must be convex")

        self.__alm_mapping_f1 = mapping_f1
        self.__alm_set_c = set_c
        if set_y is not None:
            self.__alm_set_y = set_y
        else:
            # Y not provided
            c = SetYCalculator(set_c)
            self.__alm_set_y = c.obtain()
        return self

    # ---------- DIMENSIONS --------------------------------------------

    def dim_decision_variables(self):
        """Number of decision variables

        :return: number of decision variables
        """
        return self.__u.size(1)

    def dim_parameters(self):
        """Number of parameters"""
        return self.__p.size(1)

    def dim_constraints_penalty(self):
        """Number of penalty-type constraints"""
        return 0 if self.__penalty_mapping_f2 is None \
            else self.__penalty_mapping_f2.size(1)

    def dim_constraints_aug_lagrangian(self):
        """Not implemented yet"""
        return 0 if self.__alm_mapping_f1 is None \
            else self.__alm_mapping_f1.size(1)

    # ---------- OTHER GETTERS -----------------------------------------

    @property
    def cost_function(self):
        """Cost function as a CaADi symbol"""
        return self.__cost

    @property
    def penalty_mapping_f2(self):
        """Penalty-type mapping F2 as a CasADi symbol"""
        return self.__penalty_mapping_f2

    @property
    def penalty_mapping_f1(self):
        """ALM mapping F1 as a CasADi symbol"""
        return self.__alm_mapping_f1

    @property
    def alm_set_c(self):
        """Set C in the definition of constraints: F1(u, p) in C"""
        return self.__alm_set_c

    @property
    def alm_set_y(self):
        """Set Y for the Lagrange multipliers"""
        return self.__alm_set_y

    @property
    def penalty_function(self):
        """Penalty function, g"""
        return self.__penalty_function

    @property
    def constraints(self):
        """Hard constraints; set U"""
        return self.__u_constraints

    @property
    def decision_variables(self):
        """Decision variables (CasADi symbol)"""
        return self.__u

    @property
    def parameter_variables(self):
        """Parameter variables (CasADi symbol)"""
        return self.__p



import re  # regular expressions


class OptimizerMeta:
    """Metadata of auto-generated parametric optimizer

    General metadata for the auto-generated optimizer

    The most important piece of information is the name of the
    optimizer. The optimizer will be stored in a namesake folder
    inside the target build directory.

    """

    def __init__(self,
                 optimizer_name='open_optimizer',
                 optimizer_version='0.0.0',
                 optimizer_licence='MIT',
                 optimizer_authors=['John Smith']):
        """Constructor of OptimizerMeta

        :param optimizer_name:     optimizer name (default: `open_optimizer`)
        :param optimizer_version:  version (default: `0.0.0`)
        :param optimizer_licence:  licence name or URL (default: `MIT`)
        :param optimizer_authors:  list of authors, as list of strings (default: `["John Smith"]`)

        :returns: The current instance of OptimizerMeta

        Examples:
            >>> import opengen as og
            >>> meta = og.config.OptimizerMeta()                \
            >>>     .with_version("0.0.2")                      \
            >>>     .with_authors(["P. Sopasakis", "E. Fresk"]) \
            >>>     .with_licence("CC4.0-By")                   \
            >>>     .with_optimizer_name("wow_optimizer")

        """
        self.__optimizer_name = optimizer_name
        self.__optimizer_version = optimizer_version
        self.__optimizer_licence = optimizer_licence
        self.__optimizer_author_list = optimizer_authors

    def with_version(self, optimizer_version):
        """Specify version

        Specify the version of the auto-generated optimizer.

        :param optimizer_version: version of auto-generated optimizer

        :returns: The current instance of OptimizerMeta
        """
        self.__optimizer_version = optimizer_version
        return self

    def with_authors(self, optimizer_authors):
        """Specify list of authors

        :param optimizer_authors: list of authors

        :returns: The current instance of OptimizerMeta
        """
        self.__optimizer_author_list = optimizer_authors
        return self

    def with_optimizer_name(self, optimizer_name):
        """Specify the name of the optimizer

        :param optimizer_name: name of build, may only contain letters,
        numbers and underscores, and may not start with a number

        :returns: The current instance of OptimizerMeta
        """

        if re.match(r"^[a-zA-Z_]+[\w]*$", optimizer_name):
            self.__optimizer_name = optimizer_name
            return self
        raise ValueError("invalid optimizer name")

    def with_licence(self, optimizer_licence):
        """Specify licence of auto-generated code

        :param optimizer_licence: licence name (e.g., MIT) or licence URL

        :returns: The current instance of OptimizerMeta
        """
        self.__optimizer_licence = optimizer_licence
        return self

    @property
    def optimizer_name(self):
        """Name of optimizer"""
        return self.__optimizer_name

    @property
    def version(self):
        """Version of optimizer"""
        return self.__optimizer_version

    @property
    def authors(self):
        """List of authors of optimizer"""
        return self.__optimizer_author_list

    @property
    def licence(self):
        """Licence of optimizer"""
        return self.__optimizer_licence



import re


class RosConfiguration:
    """
    Configuration of auto-generated ROS package
    """

    def __init__(self):
        """
        Constructor of an instance of RosConfiguration
        """
        self.__package_name = "open_ros"
        self.__node_name = "ros_node_optimizer"
        self.__description = "parametric optimization with OpEn"
        self.__rate = 10.0
        self.__result_topic_queue_size = 100
        self.__params_topic_queue_size = 100
        self.__publisher_subtopic = "result"
        self.__subscriber_subtopic = "parameters"

    @property
    def package_name(self):
        """
        Package name
        :return: package name (default: 'open_ros')
        """
        return self.__package_name

    @property
    def node_name(self):
        """
        Node name (default: ros_node_optimizer)
        :return:
        """
        return self.__node_name

    @property
    def publisher_subtopic(self):
        """
        Name of publisher sub-topic (default: "result")
        :return:
        """
        return self.__publisher_subtopic

    @property
    def subscriber_subtopic(self):
        """
        Name of subscriber sub-topic (default: "parameters")
        :return:
        """
        return self.__subscriber_subtopic

    @property
    def description(self):
        """
        Description of ROS package (in package.xml)
        :return: description
        """
        return self.__description

    @property
    def rate(self):
        """
        ROS node rate in Hz (default: 10)
        :return: rate
        """
        return self.__rate

    @property
    def result_topic_queue_size(self):
        """
        Size of "result" topic (default: 100)
        :return: result topic name
        """
        return self.__result_topic_queue_size

    @property
    def params_topic_queue_size(self):
        """
        Size of "parameter" topic queue (default: 100)
        :return: parameter topic name
        """
        return self.__params_topic_queue_size

    def with_package_name(self, pkg_name):
        """
        Set the package name, which is the same as the name
        of the folder that will store the auto-generated ROS node.
        The node name can contain lowercase and uppercase
        characters and underscores, but not spaces or other symbols
        :param pkg_name: package name
        :return: current object
        :raises: ValueError if pkg_name is not a legal package name
        """
        if re.match(r"^[a-zA-Z_]+[\w]*$", pkg_name):
            self.__package_name = pkg_name
            return self
        raise ValueError("invalid package name")

    def with_node_name(self, node_name):
        """
        Set the node name. The node name can contain lowercase
        and uppercase characters and underscores, but not spaces
        or other symbols
        :param node_name:
        :return: current object
        :raises: ValueError if node_name is not a legal node name
        """
        if re.match(r"^[a-zA-Z_]+[\w]*$", node_name):
            self.__node_name = node_name
            return self
        raise ValueError("invalid node name")

    def with_rate(self, rate):
        """
        Set the rate of the ROS node
        :param rate: rate in Hz
        :return: current object
        """
        self.__rate = rate
        return self

    def with_description(self, description):
        """
        Set the description of the ROS package
        :param description: description (string)
        :return: current object
        """
        self.__description = description
        return self

    def with_queue_sizes(self,
                         result_topic_queue_size=100,
                         parameter_topic_queue_size=100):
        """
        Set queue sizes for ROS node
        :param result_topic_queue_size:
        :param parameter_topic_queue_size:
        :return: current object
        """
        self.__result_topic_queue_size = result_topic_queue_size
        self.__params_topic_queue_size = parameter_topic_queue_size
        return self

    def with_publisher_subtopic(self, publisher_subtopic):
        """
        The auto-generated node will output its results to the topic
        `~/{publisher_subtopic}`. The subtopic (publisher_subtopic)
        can be specified using this method. The default subtopic name
        is 'result'. This can be configured after the package is
        generated, in `config/open_params.yaml`.

        :param publisher_subtopic: publisher sub-topic name
        :return: current object
        """
        self.__publisher_subtopic = publisher_subtopic
        return self

    def with_subscriber_subtopic(self, subscriber_subtopic):
        """
        The auto-generated node will listen for input at
        `~/{subscriber_subtopic}`. The subtopic (subscriber_subtopic)
        can be specified using this method. The default subtopic name
        is 'parameters'. This can be configured after the package is
        generated, in `config/open_params.yaml`.

        :param subscriber_subtopic: subscriber sub-topic name
        :return: :return: current object
        """
        self.__subscriber_subtopic = subscriber_subtopic
        return self



from .constraint import Constraint
from .simplex import Simplex
import numpy as np


class Ball1(Constraint):
    """Ball1 aka Norm-1 Ball

    Ball-1 with given radius.

    """

    def __init__(self, center=None, radius: float = 1.0, ):  # unless specified, radius=1
        """Constructor for a Ball1

        Args:
            :param radius: ball radius (default: 1)

        Returns:
            :return: New instance of Ball1 with given radius
        """
        if radius <= 0:
            raise ValueError("radius must be a positive number")

        if center is not None and not isinstance(center, (list, np.ndarray)):
            raise Exception("center is neither None nor a list nor np.ndarray")

        self.__radius = float(radius)
        self.__simplex = Simplex(radius)
        self.__center = None if center is None else np.array(
            [float(xi) for xi in center])

    @property
    def center(self):
        """Returns the center of the ball"""
        return self.__center

    @property
    def radius(self):
        """Returns the radius of this ball"""
        return self.__radius

    def distance_squared(self, u):
        raise NotImplementedError()

    def __proj_on_ball_centered_at_origin(self, u):
        u = np.array(u)
        if np.linalg.norm(u, 1) <= self.radius:
            return u

        n = len(u)
        x = self.__simplex.project(abs(u))
        z = np.zeros(n)
        for i, (ui, xi) in enumerate(zip(u, x)):
            z[i] = np.sign(ui) * xi
        return z

    def project(self, u):
        """Project on the current Ball-1

        Args:
            :param u: vector u

        Returns:
            :return: projection of u onto the current ball-1

        """
        if self.center is None:
            return self.__proj_on_ball_centered_at_origin(u)

        return self.__proj_on_ball_centered_at_origin(u - self.center) + self.center

    def is_convex(self):
        return True

    def is_compact(self):
        return True



import yaml
import os
import subprocess
import socket
import json
import logging
import time
import math
import pkg_resources
from threading import Thread
from retry import retry
from .solver_response import SolverResponse


class OptimizerTcpManager:
    """Client for TCP interface of parametric optimizers

    This class is used to start and stop a TCP server, which
    has been generated by <code>opengen</code>.
    """

    def __init__(self, optimizer_path=None, ip=None, port=None):
        """
        Constructs instance of <code>OptimizerTcpManager</code>

        There are three ways to use this constructor:

        - OptimizerTcpManager(optimizer_path): creates a TCP manager for a local
          TCP server using the default IP and port of that TCP server (specified
          upon code generation)
        - OptimizerTcpManager(optimizer_path, ip, port): creates a TCP manager
          for a local TCP server, but overrides the default IP and port. This way
          the user can set the address '0.0.0.0', so that the TCP server binds on
          all IPs, or '127.0.0.1' so that it is accessible only locally, or a VPN
          IP address, so that the optimizer is accessible only over a private
          network.
        - OptimizerTcpManager(ip, port): If a path is not provided, then the
          TCP manager can be used to connect to a remote TCP server, as a client,
          but cannot be used to start the server.

        Args:
            :param optimizer_path:
            path to auto-generated optimizer (just to be clear: this is
            the folder that contains <code>optimizer.yml</code>)

            :param ip:
            the user can provide the IP of a remote TCP server (must be up and
            running) so as to establish a remote connection. In that case `path`
            must be equal to `None` (see examples above)

            :param port: see ip

        Returns:
            New instance of <code>OptimizerTcpManager</code>
        """
        self.__optimizer_path = optimizer_path
        if optimizer_path is not None:
            self.__optimizer_details = None  # create attribute (including IP and port)
            self.__load_tcp_details()
            if ip is not None:
                self.__optimizer_details['tcp']['ip'] = ip
            if port is not None:
                self.__optimizer_details['tcp']['port'] = port

            # Check whether the optimizer was built with the current version of opengen
            # We can only check the optimizer version if the optimizer runs locally
            opengen_version = self.__optimizer_details['build']['opengen_version']
            current_opengen_version = pkg_resources.require("opengen")[0].version
            if current_opengen_version != opengen_version:
                logging.warn(
                    'the target optimizer was build with a different version of opengen (%s)' % opengen_version)
                logging.warn('you are running opengen version %s' % current_opengen_version)

        elif ip is not None and port is not None:
            self.__optimizer_details = {"tcp": {"ip": ip, "port": port}}
        else:
            # If the optimizer path has not been provided, both the IP and and the
            # port must be provided, otherwise, raise an exception
            raise Exception("Illegal arguments")

        logging.info("TCP/IP details: %s:%d",
                     self.__optimizer_details['tcp']['ip'],
                     self.__optimizer_details['tcp']['port'])

    def __load_tcp_details(self):
        yaml_file = os.path.join(self.__optimizer_path, "optimizer.yml")
        with open(yaml_file, 'r') as stream:
            self.__optimizer_details = yaml.safe_load(stream)

    @retry(tries=10, delay=1)
    def __obtain_socket_connection(self):
        tcp_data = self.__optimizer_details
        ip = tcp_data['tcp']['ip']
        port = tcp_data['tcp']['port']
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        s.connect((ip, port))
        return s

    def __send_receive_data(self, text_to_send, buffer_size=512, max_data_size=1048576):
        conn_socket = self.__obtain_socket_connection()
        encoded_data = text_to_send.encode()
        conn_socket.sendall(encoded_data)
        conn_socket.shutdown(socket.SHUT_WR)

        max_read_rounds = math.ceil(max_data_size/buffer_size)
        data = b''
        for _i in range(max_read_rounds):
            data_chunk = conn_socket.recv(buffer_size)
            if data_chunk is None:
                break
            data += data_chunk

        conn_socket.close()
        return data.decode()

    def ping(self):
        """Pings the server

        Pings the server to check whether it is up and running
        """
        request = '{"Ping":1}'
        data = self.__send_receive_data(request)
        return json.loads(data)

    def __check_if_server_is_running(self):
        tcp_data = self.__optimizer_details
        ip = tcp_data['tcp']['ip']
        port = tcp_data['tcp']['port']
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        return 0 == s.connect_ex((ip, port))

    @property
    def details(self):
        return self.__optimizer_details

    def start(self):
        """Starts the TCP server

        Note: this method starts a *local* server whose path must have been
        provided - we cannot start a remote server.

        The server starts on a separate thread, so this method does not block
        the execution of the caller's programme.

        """

        # Check if a path has been provided; if not,
        if self.__optimizer_path is None:
            raise Exception("No optimizer path provided - cannot start a remote server")

        # Server start data
        tcp_data = self.__optimizer_details
        ip = tcp_data['tcp']['ip']
        port = tcp_data['tcp']['port']

        # Check if any of the ip/port pairs is occupied
        if self.__check_if_server_is_running():
            msg = "Port %d not available" % port
            raise Exception(msg)

        def threaded_start():
            optimizer_details = self.__optimizer_details
            logging.info("Starting TCP/IP server at %s:%d (in a detached thread)",
                         ip, port)
            command = ['cargo', 'run', '-q']
            command += ["--release"] if optimizer_details['build']['build_mode'] == 'release' else []
            command += ['--', '--port=%d' % port, '--ip=%s' % ip]
            tcp_dir_name = "tcp_iface_" + optimizer_details['meta']['optimizer_name']
            tcp_iface_directory = os.path.join(self.__optimizer_path, tcp_dir_name)
            p = subprocess.Popen(command, cwd=tcp_iface_directory)
            p.wait()

        # start the server in a separate thread
        logging.info("Starting TCP/IP server thread")
        thread = Thread(target=threaded_start)
        thread.start()

        # ping the server until it responds so that we know it's
        # up and running
        logging.info("Waiting for server to start")
        time.sleep(0.1)
        self.ping()

    def kill(self):
        """Kills the server"""
        logging.info("Killing server")
        request = '{"Kill":1}'
        self.__send_receive_data(request)

    def call(self, p, initial_guess=None,
             initial_y=None,
             initial_penalty=None,
             buffer_len=4096,
             max_data_size=1048576) -> SolverResponse:
        """Calls the server

        Consumes the parametric optimizer by providing a parameter vector
        and, optionally, an initial guess

        Args:
             p: vector of parameters (list of float)
             initial_guess: initial guess vector (list of float)
             initial_y: initial vector of Lagrange multipliers (list of float)
             initial_penalty: initial penalty parameter (float)
             buffer_len: buffer length used to read the server response
             (default value: 4096)
             max_data_size: maximum data size that is expected to be
             received from the TCP server (default value: 1048576)

        Returns:
            Instance of SolverResponse

        """
        # Make request
        logging.debug("Sending request to TCP/IP server")
        run_message = '{"Run" : {"parameter": ['
        run_message += ','.join(map(str, p))
        run_message += ']'

        if initial_guess is not None:
            run_message += ', "initial_guess": ['
            run_message += ','.join(map(str, initial_guess))
            run_message += ']'

        if initial_y is not None:
            run_message += ', "initial_lagrange_multipliers": ['
            run_message += ','.join(map(str, initial_y))
            run_message += ']'

        if initial_penalty is not None:
            run_message += ', "initial_penalty": ' + str(float(initial_penalty))

        run_message += '}}'
        data = self.__send_receive_data(run_message, buffer_len, max_data_size)
        return SolverResponse(json.loads(data))



class Constraint:

    def distance_squared(self, u):
        raise NotImplementedError("Method `distance_squared` is not implemented")

    def project(self, u):
        raise NotImplementedError("Method `project` is not implemented")

    def is_convex(self):
        return False

    def is_compact(self):
        return False



class SolverError:
    """Class for storing solver status in the event of an error."""
    def __init__(self, error):
        """Constructs instance of <code>SolverError</code>

        Args:
            error: dictionary containing error attributes

        Returns:
            New instance of <code>SolverError</code>
        """

        for k, v in error.items():
            attribute_name = "__{}".format(k)
            setattr(self, attribute_name, v)

    @property
    def code(self):
        """
        Returns error codes:
        1000: Invalid request: Malformed or invalid JSON
        1600: Initial guess has incomplete dimensions
        1700: Wrong dimension of Lagrange multipliers
        2000: Problem solution failed (solver error)
        3003: Parameter vector has wrong length
        :return: Error code
        """
        return self.__dict__["__code"]

    @property
    def message(self):
        """
        Returns an appropriate error message matching the error code
        :return: Error message
        """
        return self.__dict__["__message"]



from ..constraints.rectangle import Rectangle
from ..constraints.ball_inf import BallInf
from ..constraints.halfspace import Halfspace


class SetYCalculator:

    LARGE_NUM = 1e12

    def __init__(self, set_c):
        self.__set_c = set_c

    def __obtain_y_with_c_compact(self):
        return BallInf(None, SetYCalculator.LARGE_NUM)

    def __obtain_y_with_c_rectangle(self):
        c = self.__set_c
        xmin = c.xmin
        xmax = c.xmax
        n = c.dimension()

        if xmin is None:
            ymin = [0.0] * n
        else:
            ymin = [-SetYCalculator.LARGE_NUM] * n
            for i in range(n):
                if xmin[i] == float('-inf'):
                    ymin[i] = 0.0

        if xmax is None:
            ymax = [0.0] * n
        else:
            ymax = [SetYCalculator.LARGE_NUM] * n
            for i in range(n):
                if xmax[i] == float('inf'):
                    ymax[i] = 0.0

        return Rectangle(ymin, ymax)

    def obtain(self):
        err_msg = 'The set C you chose is not supported yet ' \
                  '(report at https://github.com/alphaville/optimization-engine/issues)'
        if isinstance(self.__set_c, Rectangle):
            return self.__obtain_y_with_c_rectangle()
        if self.__set_c.is_compact():
            return self.__obtain_y_with_c_compact()
        if isinstance(self.__set_c, Halfspace):
            err_msg = 'For the time being, you cannot use a Halfspace in set C (issue #197)'

        raise NotImplementedError(err_msg)



import opengen as og
import unittest
import casadi.casadi as cs
import numpy as np
import math


class ConstraintsTestCase(unittest.TestCase):

    # -----------------------------------------------------------------------
    # Infinity Ball
    # -----------------------------------------------------------------------

    def test_ball_inf_origin(self):
        ball = og.constraints.BallInf(None, 1)
        x = np.array([3, 2])
        x_sym = cs.SX.sym("x", 2)
        d_num = ball.distance_squared(x)
        d_sym = float(cs.substitute(ball.distance_squared(x_sym), x_sym, x))
        self.assertAlmostEqual(d_sym, d_num, 8, "computation of distance")
        correct_squared_distance = 5.0
        self.assertAlmostEqual(d_num, correct_squared_distance,
                               8, "expected squared distance")
        # verify that it works with cs.MX
        x_sym_mx = cs.MX.sym("xmx", 2)
        sqdist_mx = ball.distance_squared(x_sym_mx)
        sqdist_mx_fun = cs.Function('sqd', [x_sym_mx], [sqdist_mx])
        self.assertAlmostEqual(correct_squared_distance,
                               sqdist_mx_fun(x)[0], 5)

    def test_ball_inf_origin_inside(self):
        ball = og.constraints.BallInf(None, 1)
        x = np.array([0.1, -0.2])
        x_sym = cs.SX.sym("x", 2)
        d_num = ball.distance_squared(x)
        d_sym = float(cs.substitute(ball.distance_squared(x_sym), x_sym, x))
        self.assertAlmostEqual(d_sym, d_num, 10, "computation of distance")
        correct_squared_distance = 0
        self.assertAlmostEqual(d_num, correct_squared_distance,
                               10, "expected squared distance")

    def test_ball_inf_xc(self):
        ball = og.constraints.BallInf([-1, -1], 0.5)
        x = np.array([1, -2])
        x_sym = cs.SX.sym("x", 2)
        d_num = ball.distance_squared(x)
        d_sym = float(cs.substitute(ball.distance_squared(x_sym), x_sym, x))
        self.assertAlmostEqual(d_sym, d_num, 8, "computation of distance")
        correct_squared_distance = 2.5
        self.assertAlmostEqual(d_num, correct_squared_distance,
                               8, "expected squared distance")

    def test_ball_inf_origin_convex(self):
        ball = og.constraints.BallInf()
        self.assertTrue(ball.is_convex())

    def test_ball_inf_origin_compact(self):
        ball = og.constraints.BallInf()
        self.assertTrue(ball.is_compact())

    # -----------------------------------------------------------------------
    # Euclidean Ball
    # -----------------------------------------------------------------------

    def test_ball_euclidean_origin(self):
        ball = og.constraints.Ball2(None, 1)
        x = np.array([2, 2])
        x_sym = cs.SX.sym("x", 2)
        d_num = ball.distance_squared(x)
        d_sym = float(cs.substitute(ball.distance_squared(x_sym), x_sym, x))
        self.assertAlmostEqual(d_sym, d_num, 8, "computation of distance")
        correct_squared_distance = 8*(1-np.sqrt(2)/4)**2
        self.assertAlmostEqual(d_sym, correct_squared_distance,
                               8, "expected squared distance")

    def test_ball_euclidean_origin_3d(self):
        ball = og.constraints.Ball2(None, 1)
        x = np.array([1, 1, 1])
        d_num = ball.distance_squared(x)
        correct_squared_distance = 0.535898384862246
        self.assertAlmostEqual(correct_squared_distance,
                               d_num, 12, "computation of distance")

    def test_ball_euclidean_origin_inside(self):
        ball = og.constraints.Ball2(None, 1)
        x = np.array([0.2, 0.8])
        x_sym = cs.SX.sym("x", 2)
        d_num = ball.distance_squared(x)
        d_sym = float(cs.substitute(ball.distance_squared(x_sym), x_sym, x))
        self.assertAlmostEqual(d_sym, d_num, 8, "computation of distance")
        correct_squared_distance = 0.0
        self.assertAlmostEqual(d_sym, correct_squared_distance,
                               8, "expected squared distance")

    def test_ball_euclidean_xc(self):
        ball = og.constraints.Ball2([1, 2], 1)
        x = [0, 1]
        x_sym = cs.SX.sym("x", 2)
        d_num = ball.distance_squared(x)
        d_sym = cs.substitute(ball.distance_squared(x_sym), x_sym, x)
        self.assertAlmostEqual(d_sym, d_num, 8, "computation of distance")
        correct_squared_distance = (np.sqrt(2) - 1) ** 2
        self.assertAlmostEqual(d_sym, correct_squared_distance,
                               8, "expected squared distance")

    def test_ball_euclidean_xc_inside(self):
        ball = og.constraints.Ball2([1, 2], 1)
        x = [1.2, 1.55]
        x_sym = cs.SX.sym("x", 2)
        d_num = ball.distance_squared(x)
        d_sym = cs.substitute(ball.distance_squared(x_sym), x_sym, x)
        self.assertAlmostEqual(d_sym, d_num, 8, "computation of distance")
        correct_squared_distance = 0.0
        self.assertAlmostEqual(d_sym, correct_squared_distance,
                               8, "expected squared distance")

    def test_ball_euclidean_origin_convex(self):
        ball = og.constraints.Ball2()
        self.assertTrue(ball.is_convex())

    def test_ball_euclidean_origin_compact(self):
        ball = og.constraints.Ball2()
        self.assertTrue(ball.is_compact())

    # -----------------------------------------------------------------------
    # Rectangle
    # -----------------------------------------------------------------------

    def test_rectangle_simple(self):
        rect = og.constraints.Rectangle([-1, -2], [4, -1])
        # some basic assertions
        self.assertListEqual([0, 1], rect.idx_bound_finite_all())
        self.assertTrue(len(rect.idx_infinite_only_xmax()) == 0)
        self.assertTrue(len(rect.idx_infinite_only_xmin()) == 0)
        self.assertEqual(2, rect.dimension())
        # squared distance
        self.assertAlmostEqual(1, rect.distance_squared([3, 0]), 8)
        self.assertAlmostEqual(4, rect.distance_squared([0, 1]), 8)
        self.assertAlmostEqual(1, rect.distance_squared([5, -1.5]), 8)
        self.assertAlmostEqual(5, rect.distance_squared([5, 1]), 8)
        # symbolic
        x_sym = cs.SX.sym("x", 2)
        d_sym = float(cs.substitute(
            rect.distance_squared(x_sym), x_sym, [5, 1]))
        self.assertAlmostEqual(5, d_sym, 8)

    def test_rectangle_pos_quant(self):
        n = 3
        rect = og.constraints.Rectangle([0.0]*n, None)
        # some basic assertions
        self.assertTrue(0 == len(rect.idx_bound_finite_all()))
        self.assertTrue(0 == len(rect.idx_infinite_only_xmin()))
        self.assertEqual([*range(n)], rect.idx_infinite_only_xmax())
        # some squared distances
        self.assertAlmostEqual(0.0, rect.distance_squared([0.0]*n), 8)
        self.assertAlmostEqual(0.0, rect.distance_squared([1.0] * n), 8)
        self.assertAlmostEqual(
            1.0, rect.distance_squared([-1.0] + [1.0] * (n-1)), 8)
        self.assertAlmostEqual(
            5.0, rect.distance_squared([-1.0, -2.0, 5.0]), 8)

    def test_rectangle_semiinf_corridor(self):
        rect = og.constraints.Rectangle([-1.0, -2.0], [float('inf'), 3.0])
        self.assertEqual([0], rect.idx_infinite_only_xmax())
        self.assertAlmostEqual(0.0, rect.distance_squared([1e16, 1.5]), 8)
        self.assertAlmostEqual(1.0, rect.distance_squared([1e16, 4.0]), 8)
        self.assertAlmostEqual(4.0, rect.distance_squared([1e16, -4.0]), 8)

    def test_rectangle_convex_i(self):
        rect = og.constraints.Rectangle([-1.0, -2.0], [1.0, 3.0])
        self.assertTrue(rect.is_convex())

    def test_rectangle_convex_ii(self):
        rect = og.constraints.Rectangle([-1.0, -2.0], [float('inf'), 3.0])
        self.assertTrue(rect.is_convex())

    def test_rectangle_compact(self):
        rect = og.constraints.Rectangle([-1.0, -2.0], [0.0, 3.0])
        self.assertTrue(rect.is_compact())

    def test_rectangle_noncompact(self):
        rect = og.constraints.Rectangle([-1.0, float('-inf')], [10.0, 3.0])
        self.assertFalse(rect.is_compact())

    # -----------------------------------------------------------------------
    # Second-Order Cone (SOC)
    # -----------------------------------------------------------------------

    def test_second_order_cone_origin_inside(self):
        soc = og.constraints.SecondOrderCone(2.0)
        # dist_C^2(0, 0, 0) = 0 [origin is in the cone]
        sq_dist = soc.distance_squared([0.0, 0.0, 0.0])
        self.assertAlmostEqual(0, sq_dist, 16)

    def test_second_order_cone_close_origin(self):
        soc = og.constraints.SecondOrderCone(2.0)
        # dist_C^2(0, 0, 0) = 0 [close-origin]
        sq_dist = soc.distance_squared([1e-12, 1e-12, 1e-12])
        self.assertAlmostEqual(0, sq_dist, 16)

    def test_second_order_cone_case_i(self):
        # dist_C^2(1, 1, 0.75) = 0 [case II]
        soc = og.constraints.SecondOrderCone(2.0)
        sq_dist = soc.distance_squared([1.0, 1.0, 0.75])
        self.assertAlmostEqual(0, sq_dist, 16)

    def test_second_order_cone_case_ii(self):
        # dist_C^2(3, 4, -11) = 146.0 [case II]
        soc = og.constraints.SecondOrderCone(2.0)
        sq_dist = soc.distance_squared([3.0, 4.0, -11.0])
        self.assertAlmostEqual(146.0, sq_dist, 16)
        sq_dist = soc.distance_squared([4.0, 3.0, -11.0])
        self.assertAlmostEqual(146.0, sq_dist, 16)

    def test_second_order_cone_case_iii(self):
        # dist_C^2(2, 3, 0.5) = 1.357... [case III]
        soc = og.constraints.SecondOrderCone(2.0)
        sq_dist = soc.distance_squared([2.0, 3.0, 0.5])
        self.assertAlmostEqual(1.35777948981440, sq_dist, 12)

    def test_second_order_cone_symbolic(self):
        soc = og.constraints.SecondOrderCone(2.0)
        u = cs.MX.sym('u', 3, 1)
        sq_dist = soc.distance_squared(u)
        u0 = [4.0, 3.0, -11.0]

        sq_dist_sx_fun = cs.Function('sqd1', [u], [sq_dist])
        self.assertAlmostEqual(146.0, sq_dist_sx_fun(u0), 16)

        sq_dist_m2 = soc.distance_squared(u)
        sq_dist_mx_fun = cs.Function('sqd2', [u], [sq_dist_m2])
        self.assertAlmostEqual(146.0, sq_dist_mx_fun(u0), 16)

    def test_second_order_cone_jacobian(self):
        soc = og.constraints.SecondOrderCone()
        # Important note: the second-order cone constraint does not work with cs.MX
        #                 An exception will be raised if the user provides an SX
        u = cs.MX.sym('u', 3)
        sq_dist = soc.distance_squared(u)
        sq_dist_jac = cs.jacobian(sq_dist, u)
        sq_dist_jac_fun = cs.Function('sq_dist_jac', [u], [sq_dist_jac])
        v = sq_dist_jac_fun([0., 0., 0.])
        for i in range(3):
            self.assertFalse(math.isnan(v[i]), "v[i] is NaN")
        self.assertAlmostEqual(0, cs.norm_2(v), 12)

    def test_second_order_cone_convex(self):
        soc = og.constraints.SecondOrderCone()
        self.assertTrue(soc.is_convex())

    def test_second_order_cone_convex(self):
        soc = og.constraints.SecondOrderCone()
        self.assertFalse(soc.is_compact())

    # -----------------------------------------------------------------------
    # No Constraints
    # -----------------------------------------------------------------------

    def test_no_constraints(self):
        whole_rn = og.constraints.NoConstraints()
        u = [1., 2., 3., 4.]
        self.assertAlmostEqual(0.0, whole_rn.distance_squared(u), 16)
        self.assertListEqual(u, whole_rn.project(u))

    def test_no_constraints_convex(self):
        whole_rn = og.constraints.NoConstraints()
        self.assertTrue(whole_rn.is_convex())

    def test_no_constraints_compact(self):
        whole_rn = og.constraints.NoConstraints()
        self.assertFalse(whole_rn.is_compact())

    # -----------------------------------------------------------------------
    # Cartesian product of constraints
    # -----------------------------------------------------------------------

    def test_cartesian(self):
        inf = float('inf')
        ball_inf = og.constraints.BallInf(None, 1)
        ball_eucl = og.constraints.Ball2(None, 1)
        rect = og.constraints.Rectangle(xmin=[0.0, 1.0, -inf, 2.0],
                                        xmax=[1.0, inf, 10.0, 10.0])
        # Segments:
        # [0, 1]
        # [2, 3, 4]
        # [5, 6, 7, 8]
        cartesian = og.constraints.CartesianProduct(
            [1, 4, 8], [ball_inf, ball_eucl, rect])
        sq_dist = cartesian.distance_squared([5, 10, 1, 1, 1, 0.5, -1, 0, 11])
        correct_sq_distance = 102.0 + (math.sqrt(3)-1.0)**2
        self.assertAlmostEqual(correct_sq_distance, sq_dist, 12)

    def test_cartesian_sx(self):
        inf = float('inf')
        ball_inf = og.constraints.BallInf(None, 1)
        ball_eucl = og.constraints.Ball2(None, 1)
        rect = og.constraints.Rectangle(xmin=[0.0, 1.0, -inf, 2.0],
                                        xmax=[1.0, inf, 10.0, 10.0])
        cartesian = og.constraints.CartesianProduct(
            [1, 4, 8], [ball_inf, ball_eucl, rect])
        u_sx = cs.SX.sym("u", 9, 1)
        _sqd_sx = cartesian.distance_squared(u_sx)
        u_mx = cs.SX.sym("u", 9, 1)
        _sqd_mx = cartesian.distance_squared(u_mx)

    def test_cartesian_segments_not_increasing(self):
        no_constraints = og.constraints.NoConstraints()
        sets = [no_constraints, no_constraints, no_constraints]
        segments = [0, 2, 2]  # should be increasing
        with self.assertRaises(ValueError) as __context:
            og.constraints.CartesianProduct(segments, sets)

    def test_cartesian_segments_negative_elements(self):
        no_constraints = og.constraints.NoConstraints()
        sets = [no_constraints, no_constraints]
        segments = [-1, 2]  # -1 is not allowed
        with self.assertRaises(ValueError) as __context:
            og.constraints.CartesianProduct(segments, sets)

    def test_cartesian_segments_different_lengths(self):
        no_constraints = og.constraints.NoConstraints()
        sets = [no_constraints, no_constraints]
        segments = [0, 2, 4]  # 3 elements (but sets has two elements)
        with self.assertRaises(ValueError) as __context:
            og.constraints.CartesianProduct(segments, sets)

    def test_cartesian_segments_empty_args(self):
        no_constraints = og.constraints.NoConstraints()
        sets = [no_constraints, no_constraints]
        with self.assertRaises(ValueError) as __context:
            og.constraints.CartesianProduct([], sets)

    # -----------------------------------------------------------------------
    # Finite Set
    # -----------------------------------------------------------------------
    def test_finite_set_dim_card(self):
        c = og.constraints.FiniteSet()
        self.assertEqual(0, c.dimension())
        self.assertEqual(0, c.cardinality())

        c = og.constraints.FiniteSet([])
        self.assertEqual(0, c.dimension())
        self.assertEqual(0, c.cardinality())

        c = og.constraints.FiniteSet([[1, 2, 3], [4, 5, 6]])
        self.assertEqual(3, c.dimension())
        self.assertEqual(2, c.cardinality())

    def test_finite_set_fail(self):
        with self.assertRaises(Exception) as __context:
            og.constraints.FiniteSet([[1., 2.], [1., 2., 3.]])

    def test_finite_set_convex(self):
        c = og.constraints.FiniteSet([[1, 2, 3]])
        self.assertTrue(c.is_convex())

    def test_finite_set_nonconvex(self):
        c = og.constraints.FiniteSet([[1, 2, 3], [4, 5, 6]])
        self.assertFalse(c.is_convex())

    def test_finite_set_compact(self):
        c = og.constraints.FiniteSet([[1, 2, 3], [4, 5, 6]])
        self.assertTrue(c.is_compact())

    # -----------------------------------------------------------------------
    # Halfspaces
    # -----------------------------------------------------------------------
    def test_halfspace_dimension(self):
        h = og.constraints.Halfspace([1, 4, 5], 1)
        self.assertEqual(3, h.dimension())

    def test_halfspace_not_compact(self):
        h = og.constraints.Halfspace([1, 4, 5], 1)
        self.assertFalse(h.is_compact())

    def test_halfspace_compact(self):
        h = og.constraints.Halfspace([0, 0, 0], -1)
        self.assertTrue(h.is_compact())

    def test_halfspace_convex(self):
        h = og.constraints.Halfspace([1, 4, 5], 1)
        self.assertTrue(h.is_convex())

    # -----------------------------------------------------------------------
    # Set Y (from C)
    # -----------------------------------------------------------------------

    # -----------------------------------------------------------------------
    # Simplex
    # -----------------------------------------------------------------------

    def test_simplex_projection(self):
        simplex = og.constraints.Simplex(alpha=2)
        y = [1, 8, 0, -4]
        z = simplex.project(y)
        self.assertAlmostEqual(
            sum(z), 2, 12, "Simplex projection sum not equal to alpha")

    def test_simplex_projection_random_spam(self):
        simplex = og.constraints.Simplex(alpha=2)
        n = 10
        for _ in range(5000):
            x = np.random.uniform(low=-100, high=100, size=n)
            alpha = np.random.uniform(low=1e-4, high=100)
            simplex = og.constraints.Simplex(alpha)
            z = simplex.project(x)
            self.assertAlmostEqual(
                sum(z), alpha, 10, "Simplex projection sum not equal to alpha")
            self.assertTrue(min(z) >= -1e-12, "Simplex projection is negative")

    def test_simplex_projection_random_optimality(self):
        # According to the projection theorem, x_star is the projection of z on a set C
        # iff <x - x_star, z - x_star> <= 0, for all x in C. Here we are testing whether
        # this holds for all x which are extreme points of C.
        for n in range(5, 60, 5):
            for _ in range(10*n):
                z = np.random.uniform(low=-100, high=100, size=n)
                alpha = np.random.uniform(low=1e-4, high=100)
                simplex = og.constraints.Simplex(alpha)
                x_star = simplex.project(z)
                # test optimality conditions:
                for j in range(n):
                    x = np.zeros((n,))
                    x[j] = alpha
                    self.assertLessEqual(
                        np.dot(x-x_star, z-x_star), 1e-10, "Simplex optimality conditions failed")

    # -----------------------------------------------------------------------
    # Ball1
    # -----------------------------------------------------------------------

    def test_ball1_project_inside_points(self):
        n = 5
        ball1 = og.constraints.Ball1(radius=1)
        for _ in range(100):
            x = np.random.uniform(low=-1, high=1, size=n)
            x /= sum(abs(x))
            x *= 1 - 1e-10
            x_star = ball1.project(x)
            self.assertLessEqual(np.linalg.norm(x - x_star, np.inf), 1e-10)

    def test_ball1_project_random_points(self):
        n = 5
        for _ in range(5000):
            r = np.random.uniform(low=1e-3, high=10)
            ball1 = og.constraints.Ball1(
                radius=r)
            x = np.random.uniform(low=-50, high=50, size=n)
            x_star = ball1.project(x)
            # Check whether the projection is inside the set
            self.assertLessEqual(np.linalg.norm(x_star, 1), r * (1 + 1e-10))
            # Check the optimality conditions
            for j in range(n):
                e = np.zeros((n,))
                e[j] = r
                self.assertLessEqual(
                    np.dot(e-x_star, x-x_star), 1e-10, "Ball1 optimality conditions failed (1)")
                e[j] = -r
                self.assertLessEqual(
                    np.dot(e-x_star, x-x_star), 1e-10, "Ball1 optimality conditions failed (2)")

    def test_ball1_project_random_points_center(self):
        n = 5
        for _ in range(5000):
            xc = np.random.uniform(low=-20, high=20, size=n)
            r = np.random.uniform(low=1e-3, high=10)
            ball1 = og.constraints.Ball1(xc, r)
            x = np.random.uniform(low=-50, high=50, size=n)
            x_star = ball1.project(x)
            # Check whether the projection is inside the set
            self.assertLessEqual(np.linalg.norm(x_star - xc, 1),
                                 r * (1 + 1e-10))
            # Check the optimality conditions
            for j in range(n):
                e = ball1.center.copy()
                e[j] += r
                self.assertLessEqual(
                    np.dot(e-x_star, x-x_star), 1e-10, "Ball1 optimality conditions failed (1)")
                e[j] -= 2*r
                self.assertLessEqual(
                    np.dot(e-x_star, x-x_star), 1e-10, "Ball1 optimality conditions failed (2)")


if __name__ == '__main__':
    unittest.main()



import opengen.definitions as og_dfn

import os
import logging
import jinja2
import shutil
import datetime

_ROS_PREFIX = 'ros_node_'


def make_dir_if_not_exists(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


def get_template(name):
    file_loader = jinja2.FileSystemLoader(og_dfn.templates_dir())
    env = jinja2.Environment(loader=file_loader, autoescape=True)
    return env.get_template(name)


def get_ros_template(name):
    file_loader = jinja2.FileSystemLoader(og_dfn.templates_subdir('ros'))
    env = jinja2.Environment(loader=file_loader, autoescape=True)
    return env.get_template(name)


class RosBuilder:
    """
    Code generation for ROS-related files
    """
    def __init__(self, meta, build_config, solver_config):
        self.__meta = meta
        self.__build_config = build_config
        self.__solver_config = solver_config
        self.__logger = logging.getLogger('opengen.builder.RosBuilder')
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(1)
        c_format = logging.Formatter('[%(levelname)s] <<ROS>> %(message)s')
        stream_handler.setFormatter(c_format)
        self.__logger.setLevel(1)
        self.__logger.addHandler(stream_handler)

    def __target_dir(self):
        return os.path.abspath(
            os.path.join(
                self.__build_config.build_dir,
                self.__meta.optimizer_name))

    def __ros_target_dir(self):
        ros_config = self.__build_config.ros_config
        ros_target_dir_name = ros_config.package_name
        return os.path.abspath(
            os.path.join(
                self.__build_config.build_dir,
                self.__meta.optimizer_name, ros_target_dir_name))

    def __generate_ros_dir_structure(self):
        self.__logger.info("Generating directory structure")
        target_ros_dir = self.__ros_target_dir()
        make_dir_if_not_exists(target_ros_dir)
        make_dir_if_not_exists(os.path.abspath(os.path.join(target_ros_dir, 'include')))
        make_dir_if_not_exists(os.path.abspath(os.path.join(target_ros_dir, 'extern_lib')))
        make_dir_if_not_exists(os.path.abspath(os.path.join(target_ros_dir, 'src')))
        make_dir_if_not_exists(os.path.abspath(os.path.join(target_ros_dir, 'msg')))
        make_dir_if_not_exists(os.path.abspath(os.path.join(target_ros_dir, 'config')))
        make_dir_if_not_exists(os.path.abspath(os.path.join(target_ros_dir, 'launch')))

    def __generate_ros_package_xml(self):
        self.__logger.info("Generating package.xml")
        target_ros_dir = self.__ros_target_dir()
        template = get_ros_template('package.xml')
        output_template = template.render(meta=self.__meta, ros=self.__build_config.ros_config)
        target_rospkg_path = os.path.join(target_ros_dir, "package.xml")
        with open(target_rospkg_path, "w") as fh:
            fh.write(output_template)

    def __generate_ros_cmakelists(self):
        self.__logger.info("Generating CMakeLists")
        target_ros_dir = self.__ros_target_dir()
        template = get_ros_template('CMakeLists.txt')
        output_template = template.render(meta=self.__meta,
                                          ros=self.__build_config.ros_config)
        target_rospkg_path = os.path.join(target_ros_dir, "CMakeLists.txt")
        with open(target_rospkg_path, "w") as fh:
            fh.write(output_template)

    def __copy__ros_files(self):
        self.__logger.info("Copying external dependencies")
        # 1. --- copy header file
        target_ros_dir = self.__ros_target_dir()
        header_file_name = self.__meta.optimizer_name + '_bindings.hpp'
        target_include_filename = os.path.abspath(
            os.path.join(
                target_ros_dir, 'include', header_file_name))
        original_include_file = os.path.abspath(
            os.path.join(self.__target_dir(), header_file_name))
        shutil.copyfile(original_include_file, target_include_filename)

        # 2. --- copy library file
        lib_file_name = 'lib' + self.__meta.optimizer_name + '.a'
        target_lib_file_name = \
            os.path.abspath(
                os.path.join(
                    target_ros_dir, 'extern_lib', lib_file_name))
        original_lib_file = os.path.abspath(
            os.path.join(
                self.__target_dir(),
                'target',
                self.__build_config.build_mode,
                lib_file_name))
        shutil.copyfile(original_lib_file, target_lib_file_name)

        # 3. --- copy msg file OptimizationParameters.msg
        original_params_msg = os.path.abspath(
            os.path.join(
                og_dfn.templates_dir(), 'ros', 'OptimizationParameters.msg'))
        target_params_msg = \
            os.path.abspath(
                os.path.join(
                    target_ros_dir, 'msg', 'OptimizationParameters.msg'))
        shutil.copyfile(original_params_msg, target_params_msg)

        # 4. --- copy msg file OptimizationResult.msg
        original_result_msg = os.path.abspath(
            os.path.join(
                og_dfn.templates_dir(), 'ros', 'OptimizationResult.msg'))
        target_result_msg = \
            os.path.abspath(
                os.path.join(
                    target_ros_dir, 'msg', 'OptimizationResult.msg'))
        shutil.copyfile(original_result_msg, target_result_msg)

    def __generate_ros_params_file(self):
        self.__logger.info("Generating open_params.yaml")
        target_ros_dir = self.__ros_target_dir()
        template = get_ros_template('open_params.yaml')
        output_template = template.render(meta=self.__meta,
                                          ros=self.__build_config.ros_config)
        target_yaml_fname \
            = os.path.join(target_ros_dir, "config", "open_params.yaml")
        with open(target_yaml_fname, "w") as fh:
            fh.write(output_template)

    def __generate_ros_node_header(self):
        self.__logger.info("Generating open_optimizer.hpp")
        target_ros_dir = self.__ros_target_dir()
        template = get_ros_template('open_optimizer.hpp')
        output_template = template.render(meta=self.__meta,
                                          ros=self.__build_config.ros_config,
                                          solver_config=self.__solver_config)
        target_rosnode_header_path \
            = os.path.join(target_ros_dir, "include", "open_optimizer.hpp")
        with open(target_rosnode_header_path, "w") as fh:
            fh.write(output_template)

    def __generate_ros_node_cpp(self):
        self.__logger.info("Generating open_optimizer.cpp")
        target_ros_dir = self.__ros_target_dir()
        template = get_ros_template('open_optimizer.cpp')
        output_template = template.render(meta=self.__meta,
                                          ros=self.__build_config.ros_config,
                                          timestamp_created=datetime.datetime.now())
        target_rosnode_cpp_path \
            = os.path.join(target_ros_dir, "src", "open_optimizer.cpp")
        with open(target_rosnode_cpp_path, "w") as fh:
            fh.write(output_template)

    def __generate_ros_launch_file(self):
        self.__logger.info("Generating open_optimizer.launch")
        target_ros_dir = self.__ros_target_dir()
        template = get_ros_template('open_optimizer.launch')
        output_template = template.render(meta=self.__meta,
                                          ros=self.__build_config.ros_config)
        target_rosnode_launch_path \
            = os.path.join(target_ros_dir, "launch", "open_optimizer.launch")
        with open(target_rosnode_launch_path, "w") as fh:
            fh.write(output_template)

    def __generate_ros_readme_file(self):
        self.__logger.info("Generating README.md")
        target_ros_dir = self.__ros_target_dir()
        template = get_ros_template('README.md')
        output_template = template.render(
            ros=self.__build_config.ros_config)
        target_readme_path \
            = os.path.join(target_ros_dir, "README.md")
        with open(target_readme_path, "w") as fh:
            fh.write(output_template)

    def __symbolic_link_info_message(self):
        target_ros_dir = self.__ros_target_dir()
        self.__logger.info("ROS package was built successfully. Now run:")
        self.__logger.info("ln -s %s  ~/catkin_ws/src/", target_ros_dir)
        self.__logger.info("cd ~/catkin_ws/; catkin_make")

    def build(self):
        """
        Build ROS-related files
        """
        self.__generate_ros_dir_structure()   # generate necessary folders
        self.__generate_ros_package_xml()     # generate package.xml
        self.__generate_ros_cmakelists()      # generate CMakeLists.txt
        self.__copy__ros_files()              # Copy certain files
        #                                     #   - C++ bindings, library, msg
        self.__generate_ros_params_file()     # generate params file
        self.__generate_ros_node_header()     # generate node .hpp file
        self.__generate_ros_node_cpp()        # generate main node .cpp file
        self.__generate_ros_launch_file()     # generate launch file
        self.__generate_ros_readme_file()     # final touch: create README.md
        self.__symbolic_link_info_message()   # Info: create symbolic link



class SolverConfiguration:
    """Configuration of solver parameters

    """

    def __init__(self):
        """Construct an instance of solver configuration parameters

        Returns:
            New instance of SolverConfiguration

        """
        self.__tolerance = 1e-4
        self.__initial_tolerance = 1e-4
        self.__lbfgs_memory = 10
        self.__max_inner_iterations = 500
        self.__max_outer_iterations = 10
        self.__constraints_tolerance = 1e-4
        self.__initial_penalty = 1.0
        self.__penalty_weight_update_factor = 5.0
        self.__max_duration_micros = 5000000
        self.__inner_tolerance_update_factor = 0.1
        self.__sufficient_decrease_coefficient = 0.1
        self.__cbfgs_alpha = None
        self.__cbfgs_epsilon = None
        self.__cbfgs_sy_epsilon = None

    # --------- GETTERS -----------------------------

    @property
    def sufficient_decrease_coefficient(self):
        """Sufficient decrease coefficient"""
        return self.__sufficient_decrease_coefficient

    @property
    def initial_penalty(self):
        """Initial penalty"""
        return self.__initial_penalty

    @property
    def cbfgs_alpha(self):
        return self.__cbfgs_alpha

    @property
    def cbfgs_epsilon(self):
        return self.__cbfgs_epsilon

    @property
    def cbfgs_sy_epsilon(self):
        return self.__cbfgs_sy_epsilon

    @property
    def tolerance(self):
        """Tolerance of inner solver"""
        return self.__tolerance

    @property
    def initial_tolerance(self):
        """Initial tolerance of inner solver"""
        return self.__initial_tolerance

    @property
    def inner_tolerance_update_factor(self):
        """"Update factor for inner tolerance"""
        return self.__inner_tolerance_update_factor

    @property
    def lbfgs_memory(self):
        """LBFGS memory for the inner solver"""
        return self.__lbfgs_memory

    @property
    def max_inner_iterations(self):
        """Maximum number of iterations for the inner solver"""
        return self.__max_inner_iterations

    @property
    def constraints_tolerance(self):
        """Tolerance on the satisfaction of the constraints"""
        return self.__constraints_tolerance

    @property
    def max_outer_iterations(self):
        """Maximum number of iterations for the outer solver"""
        return self.__max_outer_iterations

    @property
    def penalty_weight_update_factor(self):
        """Multiplicative factor for the update of the penalty weights"""
        return self.__penalty_weight_update_factor

    @property
    def max_duration_micros(self):
        """Maximum execution time in microseconds

        Returns:
            Integer value
        """
        return self.__max_duration_micros

    # --------- SETTERS -----------------------------

    def with_sufficient_decrease_coefficient(self, sufficient_decrease_coefficient):
        """
        Specify the sufficient decrease coefficient of the algorithm

            :param sufficient_decrease_coefficient: sufficient decrease coefficient

            :returns: The current object
        """
        if sufficient_decrease_coefficient <= 0.0 or sufficient_decrease_coefficient >= 1.0:
            raise Exception("sufficient decrease coeff must be in (0,1)")
        self.__sufficient_decrease_coefficient = sufficient_decrease_coefficient
        return self

    def with_initial_penalty(self, initial_penalty):
        """Initial penalty

        :param initial_penalty: initial value of penalty

        :returns: The current object
        """
        if initial_penalty <= 0:
            raise Exception("Initial penalty must be >0")

        self.__initial_penalty = float(initial_penalty)
        return self

    def with_tolerance(self, tolerance):
        """Specify tolerance

        :raises: Exception: if tolerance <= 0

        :returns: The current object
        """
        if tolerance <= 0:
            raise Exception("The tolerance must be >0")
        self.__tolerance = float(tolerance)
        return self

    def with_initial_tolerance(self, initial_tolerance):
        """
        Specify the initial tolerance

        :param initial_tolerance: initial tolerance

        :returns: The current object
        """
        if initial_tolerance <= 0:
            raise Exception("The initial tolerance must be >0")
        self.__initial_tolerance = float(initial_tolerance)
        return self

    def with_inner_tolerance_update_factor(self, inner_tol_update_factor):
        if inner_tol_update_factor <= 0 or inner_tol_update_factor > 1:
            raise Exception("The tolerance update factor must be in [0, 1)")

        self.__inner_tolerance_update_factor = float(inner_tol_update_factor)
        return self

    def with_lbfgs_memory(self, lbfgs_memory):
        """Specify L-BFGS memory

        :raises: It is required that the L-BFGS memory is larger than or
        equal to 2, otherwise an Exception is raised

        :returns: Returns the current instance of SolverConfiguration
        """
        if lbfgs_memory < 2:
            raise Exception("The L-BFGS memory must be at least equal to 2")
        self.__lbfgs_memory = int(lbfgs_memory)
        return self

    def with_max_inner_iterations(self, max_iters):
        """Maximum number of inner iterations

        :param max_iters: maximum number of iterations

        :returns: The current object
        """
        if max_iters < 1:
            raise Exception("The maximum number of inner iterations must be at least equal to 1")
        self.__max_inner_iterations = int(max_iters)
        return self

    def with_delta_tolerance(self, constraints_tolerance):
        """Tolerance on constraint violation

        :param constraints_tolerance: tolerance delta (related to constraint
        violation)

        :return: the current object
        """
        if constraints_tolerance <= 0:
            raise Exception("The constraints tolerance must be strictly positive")
        self.__constraints_tolerance = float(constraints_tolerance)
        return self

    def with_max_outer_iterations(self, max_outer_iterations):
        """Maximum number of outer iterations

        :return: the current object
        """
        if max_outer_iterations < 1:
            raise Exception("The maximum number of outer iterations must be at least equal to 1")
        self.__max_outer_iterations = int(max_outer_iterations)
        return self

    def with_penalty_weight_update_factor(self, penalty_weight_update_factor):
        """Penalty update factor

        At every outer iteration of the penalty method, the weights are
        multiplied by this factor.

        :param penalty_weight_update_factor: penalty weight update factor

        :raises: Exception: if the update factor is less than 1.0

        :return: the current object
        """
        if penalty_weight_update_factor < 1.0:
            raise Exception("The penalty update factor needs to be >= 1")
        self.__penalty_weight_update_factor = float(penalty_weight_update_factor)
        return self

    def with_max_duration_micros(self, max_duration_micros):
        """Specify the maximum duration in microseconds (must be an integer)

        :param max_duration_micros: maximum execution duration in microseconds (integer)

        :raises: Exception: if <code>max_duration_micros</code> is less than 1

        :returns: The current object
        """
        if max_duration_micros < 1:
            raise Exception("The maximum duration (in microseconds) must be >= 1")
        self.__max_duration_micros = int(max_duration_micros)
        return self

    def with_cbfgs_parameters(self, alpha, epsilon, sy_epsilon):
        """Specify the CBFGS parameters alpha and epsilon

        :param alpha: CBFGS parameter alpha
        :param epsilon: CBFGS parameter epsilon
        :param sy_epsilon: Tolerance on the s-y inner product

        :returns: the current object
        """
        if epsilon < 0.0:
            raise Exception("CBFGS parameter epsilon must be positive")

        if alpha < 0.0:
            raise Exception("CBFGS parameter alpha must be positive")

        self.__cbfgs_epsilon = epsilon
        self.__cbfgs_alpha = alpha
        self.__cbfgs_sy_epsilon = sy_epsilon
        return self



from opengen.functions.is_symbolic import is_symbolic


def rosenbrock(u_, p_):
    """Rosenbrock functions with parameters <code>p = [a, b]</code>"""
    if not is_symbolic(p_) or p_.size()[0] != 2:
        raise Exception('illegal parameter p_ (must be SX of size (2,1))')
    if not is_symbolic(u_):
        raise Exception('illegal parameter u_ (must be SX)')
    nu = u_.size()[0]
    a = p_[0]
    b = p_[1]
    ros_fun = 0
    for i in range(nu-1):
        ros_fun += b*(u_[i+1]-u_[i]**2)**2 + (a-u_[i])**2
    return ros_fun



import casadi.casadi as cs
import numpy as np
from .is_numeric import is_numeric
from .is_symbolic import is_symbolic


def norm2_squared(u):
    if (isinstance(u, list) and all(is_numeric(x) for x in u))\
                or isinstance(u, np.ndarray):
        # if `u` is a numeric vector
        return np.dot(u, u)
    if is_symbolic(u):
        return cs.dot(u, u)
    raise Exception("Illegal argument")



from . import constraint
from typing import List


class CartesianProduct(constraint.Constraint):

    def __init__(self, segments: List[int], constraints: List[constraint.Constraint]):
        """
        Construct a Cartesian product of constraints by providing a list of sets
        and their dimensions as follows: an n-dimensional vector x can be partitioned
        into subvectors as x = (x1, x2, ..., xs), where each xi has dimension mi.
        For example consider the 5-dimensional vector x = (x_0, x_1, x_2, x_3, x_4),
        which can be partitioned into x1 = (x_0, x_1) and x2 = (x_2, x_3, x_4).
        We can associate with x1 the indices [0, 1] and with x2 the indices [2, 3, 4].
        The *segment ids* are the indices 1 and 4.

        Example:
        > ball = og.constraints.Ball2(None, 1.5)
        > rect = og.constraints.Rectangle(xmin=[-1,-2,-3], xmax=[0, 10, -1])
        > free = og.constraints.NoConstraints()
        > segment_ids = [1, 4, 9]
        > my_set = og.constraints.CartesianProduct(segment_ids, [ball, rect])

        :param segments: ids of segments
        :param constraints: list of sets
        """

        if not segments or not constraints:
            raise ValueError("segments and constraints must be nonempty lists")

        if any([segments[i] >= segments[i+1] for i in range(len(segments)-1)]):
            raise ValueError("segments should be a list of integers in strictly ascending order")

        if segments[0] < 0:
            raise ValueError("the first element of segment must be a positive integer")

        if len(segments) != len(constraints):
            raise ValueError("segments and constraints must have equal dimensions")

        self.__segments = segments
        self.__constraints = constraints

    @property
    def constraints(self):
        """

        :return: list of constraints comprising the current instance of CartesianProduct
        """
        return self.__constraints

    @property
    def segments(self):
        """
        :return: list of segments
        """
        return self.__segments

    def segment_dimension(self, i):
        """
        Dimension of segment i
        :param i: index of segment (starts at 0)
        :return: dimension of i-th index
        """
        if i == 0:
            return self.__segments[0] + 1
        return self.__segments[i] - self.__segments[i-1]

    def distance_squared(self, u):
        """
        Squared distance of given vector, u, from the current instance of
        CartesianProduct
        :param u: vector u
        :return: squared distance (float)
        """
        squared_distance = 0.0
        num_segments = len(self.__segments)
        idx_previous = -1
        for i in range(num_segments):
            idx_current = self.__segments[i]
            ui = u[idx_previous+1:idx_current+1]
            current_sq_dist = self.__constraints[i].distance_squared(ui)
            idx_previous = idx_current
            squared_distance += current_sq_dist

        return squared_distance

    def project(self, u):
        raise NotImplementedError()

    def is_convex(self):
        flag = True
        for c in self.__constraints:
            flag &= c.is_convex()
        return flag

    def is_compact(self):
        for set_i in self.__constraints:
            if not set_i.is_compact():
                return False
        return True



import sys
import os
import casadi.casadi as cs
import opengen as og

optimizers_dir = "my_optimizers"
optimizer_name = "rosenbrock"

u = cs.SX.sym("u", 5)  # decision variable (nu = 5)
p = cs.SX.sym("p", 2)  # parameter (np = 2)
phi = og.functions.rosenbrock(u, p)  # cost function

bounds = og.constraints.Halfspace([1., 2., 1., 5., 2.], -10.39)
problem = og.builder.Problem(u, p, phi).with_constraints(bounds)
meta = og.config.OptimizerMeta() \
    .with_optimizer_name(optimizer_name)

tcp_config = og.config.TcpServerConfiguration(bind_port=3305)
build_config = og.config.BuildConfiguration() \
    .with_build_directory(optimizers_dir) \
    .with_build_mode(og.config.BuildConfiguration.DEBUG_MODE) \
    .with_build_python_bindings() \
    .with_tcp_interface_config()

builder = og.builder.OpEnOptimizerBuilder(problem,
                                          meta,
                                          build_config,
                                          og.config.SolverConfiguration())
builder.build()

sys.path.insert(1, os.path.join(optimizers_dir, optimizer_name))
rosenbrock = __import__(optimizer_name)

solver = rosenbrock.solver()
result = solver.run([1., 2.])
print(result.solution)



#!/usr/bin/env python

from setuptools import setup, find_packages
import io
import os

here = os.path.abspath(os.path.dirname(__file__))

NAME = 'opengen'

# Import version from file
version_file = open(os.path.join(here, 'VERSION'))
VERSION = version_file.read().strip()

DESCRIPTION = 'Optimization Engine Code Generator'


# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=long_description,
      long_description_content_type='text/markdown',
      author=['Pantelis Sopasakis', 'Emil Fresk'],
      author_email='p.sopasakis@gmail.com',
      license='MIT License',
      packages=find_packages(
          exclude=["tests", "icasadi", "templates"]),
      include_package_data=True,
      install_requires=[
          'jinja2', 'casadi', 'pyyaml', 'retry', 'numpy'
      ],
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: MIT License',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python',
          'Programming Language :: Rust',
          'Intended Audience :: Science/Research',
          'Topic :: Software Development :: Libraries',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Topic :: Software Development :: Code Generators',
          'Topic :: Software Development :: Embedded Systems'
      ],
      keywords=['optimization', 'nonconvex', 'embedded'],
      url=(
          'https://github.com/alphaville/optimization-engine'
      ),
      zip_safe=False)



from .ball1 import *
from .ball2 import *
from .rectangle import *
from .constraint import *
from .ball_inf import *
from .soc import *
from .no_constraints import *
from .cartesian import *
from .zero import *
from .finite_set import *
from .halfspace import *
from .simplex import *
